/*
 * Copyright (C) 2006 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: bcmst.c,v 1.4 2006/10/26 19:36:08 nalin Exp $"

#include "../config.h"

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <krb5/krb5.h>

#include <cert.h>
#include <certdb.h>
#include <certt.h>
#include <cryptohi.h>
#include <pk11pub.h>

#include "certs.h"
#include "commont.h"
#include "bcmst.h"
#include "pkinit.h"

static unsigned char bcms_null_bytes[] = {SEC_ASN1_NULL, 0};
static SECItem bcms_null_item = {
	.data = &bcms_null_bytes[0],
	.len = 2,
};

const SEC_ASN1Template
bcms_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct content_info, content),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_constructed_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct content_info, content),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_encapsulated_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct content_info, content),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_signer_identifier_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct signer_identifier, signer_type),
	.sub = NULL,
	.size = sizeof(struct signer_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct signer_identifier,
			   signer_value.subject_key_identifier),
	.sub = &SEC_AnyTemplate,
	.size = signer_type_subject_key_identifier,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_identifier,
			   signer_value.issuer_and_serial_number),
	.sub = &common_issuer_and_serial_number_template,
	.size = signer_type_issuer_and_serial_number,
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_signer_infos_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct signer_infos),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct signer_infos, version),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_infos, sid),
	.sub = &bcms_signer_identifier_template,
	.size = sizeof(struct signer_identifier),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_infos, digest_algorithm),
	.sub = &common_algorithm_identifier_template,
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signer_infos, signed_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_infos, signature_algorithm),
	.sub = &common_algorithm_identifier_template,
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct signer_infos, signature),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signer_infos, unsigned_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_signed_data_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct signed_data),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct signed_data, version),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct signed_data, digest_algorithms),
	.sub = common_algorithm_identifier_template,
	.size = 0,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signed_data, encap_content_info),
	.sub = &bcms_encapsulated_content_info_template,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signed_data, certificates),
	.sub = &SEC_SetOfAnyTemplate,
	.size = sizeof(SECItem**),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signed_data, crls),
	.sub = &SEC_SetOfAnyTemplate,
	.size = sizeof(SECItem**),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct signed_data, signer_infos),
	.sub = bcms_signer_infos_template,
	.size = 0,
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
bcms_external_principal_identifiers_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct external_principal_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct external_principal_identifier,
			   der_subject_name),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct external_principal_identifier,
			   issuer_and_serial_number),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct external_principal_identifier,
			   subject_key_identifier),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
bcms_sequence_of_external_principal_identifiers_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = bcms_external_principal_identifiers_template,
	.size = sizeof(struct external_principal_identifier **),
	},
};

const SEC_ASN1Template
bcms_recipient_identifier_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct recipient_identifier,
			   recipient_identifier_type),
	.sub = NULL,
	.size = sizeof(struct recipient_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct recipient_identifier,
			   recipient_identifier_value.subject_key_identifier),
	.sub = &SEC_AnyTemplate,
	.size = recipient_identifier_type_subject_key_identifier,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct recipient_identifier,
			   recipient_identifier_value.issuer_and_serial_number),
	.sub = &common_issuer_and_serial_number_template,
	.size = recipient_identifier_type_issuer_and_serial_number,
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_encrypted_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct encrypted_content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct encrypted_content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct encrypted_content_info,
			   content_encryption_algorithm),
	.sub = &common_algorithm_identifier_template,
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct encrypted_content_info, encrypted_content),
	.sub = SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_key_trans_recipient_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct key_trans_recipient_info),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct key_trans_recipient_info, version),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct key_trans_recipient_info, rid),
	.sub = &bcms_recipient_identifier_template,
	.size = sizeof(struct recipient_identifier),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct key_trans_recipient_info,
			   key_encryption_algorithm),
	.sub = &common_algorithm_identifier_template,
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct key_trans_recipient_info, encrypted_key),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
bcms_recipient_infos_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct recipient_infos, recipient_info_type),
	.sub = NULL,
	.size = sizeof(struct recipient_infos),
	},
	{
	.kind = SEC_ASN1_CONSTRUCTED | SEC_ASN1_INLINE,
	.offset = offsetof(struct recipient_infos, recipient_info_value.ktri),
	.sub = &bcms_key_trans_recipient_info_template,
	.size = recipient_info_type_key_trans,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct recipient_infos,
			   recipient_info_value.der_kari),
	.sub = &SEC_AnyTemplate,
	.size = recipient_info_type_key_agree,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct recipient_infos,
			   recipient_info_value.der_kari),
	.sub = &SEC_AnyTemplate,
	.size = recipient_info_type_key_agree,
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
bcms_enveloped_data_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct enveloped_data),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct enveloped_data, version),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct enveloped_data, der_originator_info),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct enveloped_data, recipient_infos),
	.sub = &bcms_recipient_infos_template,
	.size = 0,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct enveloped_data, encrypted_content_info),
	.sub = &bcms_encrypted_content_info_template,
	.size = sizeof(struct encrypted_content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct enveloped_data, der_unsigned_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

struct content_info *
bcms_decode_content_info(PLArenaPool *pool, SECItem *item)
{
	struct content_info *ci;

	ci = PORT_ArenaZAlloc(pool, sizeof(*ci));
	if (ci == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ci, bcms_content_info_template,
			       item) != SECSuccess) {
		if (SEC_ASN1DecodeItem(pool, ci,
				       bcms_constructed_content_info_template,
				       item) != SECSuccess) {
			pkinit_debug("Error decoding content-info!\n");
			return NULL;
		}
	}
	return ci;
}

SECItem *
bcms_encode_content_info(PLArenaPool *pool, struct content_info *ci)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ci,
			       bcms_content_info_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding content-info!\n");
	return NULL;
}

SECItem *
bcms_encode_constructed_content_info(PLArenaPool *pool, struct content_info *ci)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ci,
			       bcms_constructed_content_info_template) ==
	    &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding (constructed) content-info!\n");
	return NULL;
}

struct signed_data *
bcms_decode_signed_data(PLArenaPool *pool, SECItem *item)
{
	struct signed_data *sd;

	sd = PORT_ArenaZAlloc(pool, sizeof(*sd));
	if (sd == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, sd, bcms_signed_data_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding signed-data!\n");
		return NULL;
	}
	return sd;
}

SECItem *
bcms_encode_signed_data(PLArenaPool *pool, struct signed_data *sd)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, sd,
			       bcms_signed_data_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding signed-data!\n");
	return NULL;
}

struct enveloped_data *
bcms_decode_enveloped_data(PLArenaPool *pool, SECItem *item)
{
	struct enveloped_data *ed;

	ed = PORT_ArenaZAlloc(pool, sizeof(*ed));
	if (ed == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ed, bcms_enveloped_data_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding enveloped-data!\n");
		return NULL;
	}
	return ed;
}

SECItem *
bcms_encode_enveloped_data(PLArenaPool *pool, struct enveloped_data *ed)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ed,
			       bcms_enveloped_data_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding enveloped-data!\n");
	return NULL;
}

struct content_info *
bcms_build_content_info(PLArenaPool *pool, SECOidTag content_type_tag,
			SECItem *content_data)
{
	SECOidData *oid;
	oid = SECOID_FindOIDByTag(content_type_tag);
	if (oid != NULL) {
		return bcms_build_content_info_oid(pool, &oid->oid,
						   content_data);
	}
	return NULL;
}

struct content_info *
bcms_build_content_info_oid(PLArenaPool *pool, SECItem *content_type_oid,
			    SECItem *content_data)
{
	struct content_info *ret;
	SECOidData *oid;
	if (content_type_oid == NULL) {
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
		if (oid == NULL) {
			return NULL;
		}
		content_type_oid = &oid->oid;
	}
	ret = PORT_ArenaZAlloc(pool, sizeof(*ret));
	if (ret != NULL) {
		ret->content_type = *content_type_oid;
		ret->content = *content_data;
	}
	return NULL;
}

/* Turn a digest or signature tag into the corresponding digest tag. */
static SECOidTag
sanify_digtag(SECOidTag digtag)
{
	switch (digtag) {
	case SEC_OID_PKCS1_MD5_WITH_RSA_ENCRYPTION:
	case SEC_OID_MD5:
		digtag = SEC_OID_MD5;
		break;
	case SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION:
	case SEC_OID_ANSIX9_DSA_SIGNATURE_WITH_SHA1_DIGEST:
	case SEC_OID_SHA1:
		digtag = SEC_OID_SHA1;
		break;
	case SEC_OID_PKCS1_SHA256_WITH_RSA_ENCRYPTION:
	case SEC_OID_SHA256:
		digtag = SEC_OID_SHA256;
		break;
	case SEC_OID_PKCS1_SHA384_WITH_RSA_ENCRYPTION:
	case SEC_OID_SHA384:
		digtag = SEC_OID_SHA384;
		break;
	case SEC_OID_PKCS1_SHA512_WITH_RSA_ENCRYPTION:
	case SEC_OID_SHA512:
		digtag = SEC_OID_SHA512;
		break;
	default:
		break;
	}
	return digtag;
}

/* Map a digest/encrypt tag pair into a signature tag. */
static SECOidData *
find_sigoid(SECOidTag digtag, SECOidTag enctag)
{
	unsigned int i;
	const struct {
		SECOidTag digtag, enctag, sigtag;
	} tagmap[] = {
		{SEC_OID_MD5, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_MD5_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA1, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA256, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA256_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA384, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA384_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA512, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA512_WITH_RSA_ENCRYPTION},
	};
	for (i = 0; i < sizeof(tagmap) / sizeof(tagmap[0]); i++) {
		if ((tagmap[i].digtag == digtag) &&
		    (tagmap[i].enctag == enctag)) {
			return SECOID_FindOIDByTag(tagmap[i].sigtag);
		}
	}
	return NULL;
}

/* Find the certificate in the list which matches the signer_identifier. */
static CERTCertificate *
find_cert_from_sid(CERTCertDBHandle *certdb, SECItem **certificates,
		   struct signer_identifier *sid)
{
	CERTCertificate *cert, **certs;
	struct issuer_and_serial_number *is;
	SECItem *spki;
	int i, n_certificates;

	/* First, check that we actually have certificates. */
	for (i = 0; (certificates != NULL) && (certificates[i] != NULL); i++) {
		continue;
	}
	if (i == 0) {
		pkinit_debug("Could not find certificate corresponding to "
			     "signer info in an empty certificate list!\n");
		return NULL;
	}
	n_certificates = i;
	certs = NULL;

	/* Parse the list of certificates so that we can examine them. */
	if (CERT_ImportCerts(certdb, certUsageSSLClient,
			     n_certificates,
			     certificates,
			     &certs,
			     PR_FALSE,
			     PR_FALSE,
			     NULL) != SECSuccess) {
		pkinit_debug("Error importing certificates!\n");
		return NULL;
	}

	/* Walk the list of certificates. */
	for (i = 0; i < n_certificates; i++) {
		cert = certs[i];
		switch (sid->signer_type) {
		case signer_type_issuer_and_serial_number:
			/* The signer is identified by issuer/serial. */
			is = &sid->signer_value.issuer_and_serial_number;
			if (SECITEM_ItemsAreEqual(&is->issuer,
						  &cert->derIssuer) &&
			    SECITEM_ItemsAreEqual(&is->serial,
						  &cert->serialNumber)) {
				return cert;
			}
			break;
		case signer_type_subject_key_identifier:
			/* The signer is identified by the key identifier. */
			spki = &sid->signer_value.subject_key_identifier;
			if (SECITEM_ItemsAreEqual(spki, &cert->subjectKeyID)) {
				return cert;
			}
			break;
		default:
			break;
		}
	}
	pkinit_debug("Could not find certificate in list which corresponded "
		     "to signer_info!\n");
	return NULL;
}

/* Verify ALL of the signatures on a signed_data, and if it has exactly one
 * signer, return the certificate of that signer. */
SECStatus
bcms_verify_signed_data(CERTCertDBHandle *certdb,
			struct signed_data *signed_data,
			CERTCertificate **lone_signer)
{
	SECOidData *digoid, *encoid, *sigoid;
	SECOidTag digtag;
	CERTCertificate *cert;
	SECKEYPublicKey *public;
	struct signer_infos **signer_infos;
	struct algorithm_identifier *algorithm;
	SECItem *econtent;
	SECStatus status;
	int i;

	signer_infos = signed_data->signer_infos;
	status = SECSuccess;
	*lone_signer = NULL;
	cert = NULL;
	for (i = 0; (signer_infos != NULL) && (signer_infos[i] != NULL); i++) {
		/* Pick out the digest/encryption/signature types. */
		pkinit_debug("Signature %d:\n", i);
		algorithm = &signer_infos[i]->signature_algorithm;
		encoid = SECOID_FindOID(&algorithm->algorithm);
		algorithm = &signer_infos[i]->digest_algorithm;
		digoid = SECOID_FindOID(&algorithm->algorithm);
		if (digoid != NULL) {
			digtag = sanify_digtag(digoid->offset);
			if (SECOID_FindOIDByTag(digtag)) {
				digoid = SECOID_FindOIDByTag(digtag);
			}
		}
		sigoid = NULL;
		if (digoid && encoid) {
			pkinit_debug("\tencryption: %s\n", encoid->desc);
			pkinit_debug("\tdigest: %s\n", digoid->desc);
			sigoid = find_sigoid(digoid->offset, encoid->offset);
			if (sigoid == NULL) {
				pkinit_debug("\tunknown signature type!\n");
				break;
			}
		} else {
			pkinit_debug("\tunknown signature or encryption "
				     "type!\n");
			break;
		}
		/* Find the signing certificate in the list. */
		pkinit_debug("\tsignature: %s\n", sigoid->desc);
		cert = find_cert_from_sid(certdb, signed_data->certificates,
					  &signer_infos[i]->sid);
		/* Extract the verification key from the certificate. */
		public = NULL;
		if (cert != NULL) {
			pkinit_debug("Found signing certificate.\n");
			public = CERT_ExtractPublicKey(cert);
		} else {
			pkinit_debug("Signing certificate not included.\n");
			/* No/invalid cert? Bail. */
			break;
		}
		econtent = NULL;
		if (public != NULL) {
			pkinit_debug("Pulled out public key.\n");
			econtent = &signed_data->encap_content_info.content;
		} else {
			/* No/invalid public key? Bail. */
			pkinit_debug("Unable to extract key from cert!\n");
			break;
		}
		/* Verify the signature. */
		if (econtent != NULL) {
			status = VFY_VerifyData(econtent->data,
						econtent->len,
						public,
						&signer_infos[i]->signature,
						sigoid->offset,
						NULL);
			if (status == SECSuccess) {
				pkinit_debug("Signature verified.\n");
			} else {
				break;
			}
		} else {
			/* No data to verify the signature? Bail. */
			break;
		}
	}
	/* If we had no signatures, we've failed. */
	if (i == 0) {
		pkinit_debug("No signatures to verify!\n");
		return SECFailure;
	}
	/* If we had only one signer, return a copy of the signer's
	 * certificate. */
	if ((i == 1) && (lone_signer != NULL)) {
		*lone_signer = cert;
	}
	/* If we had failed any check (meaning we didn't run out of signatures
	 * to verify), we've failed. */
	if (signer_infos[i] == NULL) {
		return SECSuccess;
	}
	/* Return the status of the last-attempted signature verification. */
	return status;
}

/* Make a copy of the list of certificates, ensuring that the passed-in value
 * is in there somewhere. */
static SECItem **
bcms_make_certificate_list(PLArenaPool *pool, SECItem **list,
			   CERTCertificate *cert)
{
	PRBool found;
	SECItem **nlist;
	int i;

	found = PR_FALSE;
	for (i = 0; (list != NULL) && (list[i] != NULL) && !found; i++) {
		if (SECITEM_ItemsAreEqual(list[i], &cert->derCert)) {
			found = PR_TRUE;
			break;
		}
	}

	if (found) {
		nlist = list;
	} else {
		/* i contains the length of the current list */
		nlist = PORT_ArenaZAlloc(pool, sizeof(SECItem*) * (i + 2));
		if (nlist == NULL) {
			pkinit_debug("Out of memory\n");
			return NULL;
		}
		/* add our entry to the end of the list */
		nlist[i] = &cert->derCert;
		/* copy the rest of the list */
		while (--i >= 0) {
			nlist[i] = list[i];
		}
	}

	return nlist;
}

/* Add a signature to the signed_data structure. */
static SECStatus
bcms_add_signer_to_signed_data(PLArenaPool *pool,
			       struct signed_data *sig,
			       SECOidTag digtag,
			       SECOidTag enctag,
			       PK11SlotInfo *preferred_slot,
			       CERTCertificate *cert,
			       struct pkinit_pwcb_args *pwcb_args)
{
	SECOidData *digoid, *encoid, *sigoid;
	SECKEYPrivateKey *privkey;
	struct signer_infos **signer_infos, *signer_info;
	SECItem **certificates;
	struct algorithm_identifier **digest_algorithms;
	SECStatus status;
	SECItem *econtent;
	SECItem *null_item;
	int signer_count, i;

	/* Double-check that we actually have something to sign. */
	if (sig->encap_content_info.content.len == 0) {
		pkinit_debug("No content to sign!\n");
		return SECFailure;
	}

	/* Find the private key which corresponds to the certificate. */
	privkey = NULL;
	if (preferred_slot != NULL) {
		privkey = PK11_FindPrivateKeyFromCert(preferred_slot, cert,
						      pwcb_args);
	}
	if (privkey == NULL) {
		privkey = PK11_FindKeyByAnyCert(cert, pwcb_args);
	}
	if (privkey == NULL) {
		pkinit_debug("Unable to load private key.\n");
		return SECFailure;
	}

	/* Create the new signer_info structure and set its version. */
	signer_info = PORT_ArenaZAlloc(pool, sizeof(*signer_info));
	if (SEC_ASN1EncodeInteger(pool, &signer_info->version, 1) !=
	    &signer_info->version) {
		pkinit_debug("Error encoding version.\n");
		return SECFailure;
	}

	/* Set up the signer identifier. */
	signer_info->sid.signer_type =
		signer_type_issuer_and_serial_number;
	signer_info->sid.signer_value.issuer_and_serial_number.issuer =
		cert->derIssuer;
	signer_info->sid.signer_value.issuer_and_serial_number.serial =
		cert->serialNumber;

	/* Set up the digest algorithm. */
	digoid = SECOID_FindOIDByTag(digtag);
	if (digoid == NULL) {
		pkinit_debug("Invalid digest type.\n");
		return SECFailure;
	}
	signer_info->digest_algorithm.algorithm = digoid->oid;
	null_item = SECITEM_ArenaDupItem(pool, &bcms_null_item);
	signer_info->digest_algorithm.parameters = null_item;

	/* No authenticated attributes. */
	memset(&signer_info->signed_attrs, 0,
	       sizeof(signer_info->signed_attrs));

	/* Set up the signing algorithm. */
	sigoid = find_sigoid(sanify_digtag(digtag), enctag);
	if (sigoid == NULL) {
		pkinit_debug("Could not determine signature type.\n");
		return SECFailure;
	} else {
		pkinit_debug("Generating signature: %s.\n", sigoid->desc);
	}

	encoid = SECOID_FindOIDByTag(enctag);
	if (encoid == NULL) {
		pkinit_debug("Could not determine encryption type.\n");
		return SECFailure;
	}
	signer_info->signature_algorithm.algorithm = encoid->oid;
	signer_info->signature_algorithm.parameters = null_item;

	/* Create a PKCS1 signature using the selected digest. */
	sigoid = find_sigoid(sanify_digtag(digtag), enctag);
	econtent = &sig->encap_content_info.content;
	status = SEC_SignData(&signer_info->signature,
			      econtent->data,
			      econtent->len,
			      privkey,
			      sigoid->offset);

	if (status != SECSuccess) {
		pkinit_debug("Error signing data.\n");
		return SECFailure;
	}

	/* No unauthenticated attributes. */
	memset(&signer_info->unsigned_attrs, 0,
	       sizeof(signer_info->unsigned_attrs));

	/* Append our signer_info to the existing list.  Start by counting the
	 * existing list's size. */
	signer_infos = sig->signer_infos;
	for (signer_count = 0;
	     (signer_infos != NULL) && (signer_infos[signer_count] != NULL);
	     signer_count++) {
		continue;
	}
	signer_infos = PORT_ArenaZAlloc(pool,
					sizeof(struct signer_info*) *
					(signer_count + 2));
	if (signer_infos == NULL) {
		pkinit_debug("Out of memory\n");
		return SECFailure;
	}

	/* Save the old elements of the list, and then append the new one. */
	for (i = 0; i < signer_count; i++) {
		signer_infos[i] = sig->signer_infos[i];
	}
	signer_infos[i] = signer_info;

	/* Save the digest algorithm in the signed_data's list. */
	digest_algorithms = common_make_algorithm_identifier_list(pool,
								  sig->digest_algorithms,
								  &signer_info->digest_algorithm);
	if (digest_algorithms == NULL) {
		pkinit_debug("Out of memory\n");
		return SECFailure;
	}

	/* Save the certificate to the signed_data's list. */
	certificates = bcms_make_certificate_list(pool, sig->certificates,
						  cert);
	if (certificates == NULL) {
		pkinit_debug("Out of memory\n");
		return SECFailure;
	}

	/* Save the new lists. */
	sig->digest_algorithms = digest_algorithms;
	sig->signer_infos = signer_infos;
	sig->certificates = certificates;

	return status;
}

/* Add a certificiate to the signed_data structure. */
static SECStatus
bcms_add_cert_to_signed_data(PLArenaPool *pool,
			     struct signed_data *sig,
			     CERTCertificate *cert)
{
	SECItem **certificates;
	if (cert != NULL) {
		certificates = bcms_make_certificate_list(pool,
							  sig->certificates,
							  cert);
		if (certificates == NULL) {
			return SECFailure;
		}
		sig->certificates = certificates;
	}
	return SECSuccess;
}

/* Add a certificiate's chain (minus itself and its root) to the signed_data
 * structure. */
static SECStatus
bcms_add_cert_chain_to_signed_data(PLArenaPool *pool,
				   struct signed_data *sig,
				   CERTCertDBHandle *certdb,
				   CERTCertificate *cert)
{
	CERTCertificate *issuer;
	CERTIssuerAndSN isn;

	issuer = NULL;
	pkinit_debug("Walking certifying chain for \"%s\".\n",
		     cert->nickname);
	if (cert->authKeyID != NULL) {
		issuer = CERT_FindCertBySubjectKeyID(certdb,
						     &cert->authKeyID->keyID);
	} else {
		isn.issuer = issuer->issuer;
		isn.derIssuer = issuer->derIssuer;
		isn.serialNumber = issuer->serialNumber;
		issuer = CERT_FindCertByIssuerAndSN(certdb, &isn);
	}
	while ((issuer != NULL) && (!issuer->isRoot)) {
		pkinit_debug("Adding chain link \"%s\".\n",
			     issuer->nickname);
		if (bcms_add_cert_to_signed_data(pool, sig,
						 issuer) != SECSuccess) {
			return SECFailure;
		}
		if (issuer->authKeyID != NULL) {
			issuer = CERT_FindCertBySubjectKeyID(certdb,
							     &issuer->authKeyID->keyID);
		} else {
			isn.issuer = issuer->issuer;
			isn.derIssuer = issuer->derIssuer;
			isn.serialNumber = issuer->serialNumber;
			issuer = CERT_FindCertByIssuerAndSN(certdb, &isn);
		}
	}
	if (issuer == NULL) {
		pkinit_debug("No more links in \"%s\"'s cert chain found.\n",
			     cert->nickname);
		return SECFailure;
	} else {
		if (issuer->isRoot) {
			pkinit_debug("Stopping at root \"%s\".\n",
				     issuer->nickname);
		}
	}
	return SECSuccess;
}

/* Populate a new signed_data structure with the given encapsulated content
 * type and data. */
static SECStatus
bcms_populate_signed_data(PLArenaPool *pool,
			  struct signed_data *signed_data,
			  const SECItem *encapsulated_oid,
			  const SECItem *encapsulated_data)
{
	/* Build the signed-data payload, version first. */
	if (SEC_ASN1EncodeInteger(pool, &signed_data->version, 1) !=
	    &signed_data->version) {
		pkinit_debug("Error encoding version.\n");
		return SECFailure;
	}

	/* Initialize as empty the list of digest algorithms. */
	signed_data->digest_algorithms = NULL;

	/* Build the encapsulated content-info. */
	signed_data->encap_content_info.content_type = *encapsulated_oid;
	signed_data->encap_content_info.content = *encapsulated_data;

	/* Initially, we have no signer-infos, so no certificates and stuff. */
	signed_data->certificates = NULL;
	signed_data->crls = NULL;
	signed_data->signer_infos = NULL;

	return SECSuccess;
}

/* Build a new encoded signed_data. */
SECItem *
bcms_create_signed_data(PLArenaPool *pool,
			CERTCertDBHandle *certdb,
			const SECItem *content_type,
			SECItem *data,
			CERTCertificate *signer,
			SECOidTag digest_algorithm,
			PK11SlotInfo *preferred_slot,
			struct pkinit_pwcb_args *pwcb_args)
{
	SECStatus status;
	SECOidData *oid;
	struct signed_data *signed_data;

	if (content_type == NULL) {
		/* Retrieve the OID for the default of "data". */
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
		if (oid == NULL) {
			return NULL;
		}
		content_type = &oid->oid;
	}

	/* Build the new structure. */
	signed_data = PORT_ArenaZAlloc(pool, sizeof(*signed_data));
	if (signed_data == NULL) {
		return NULL;
	}

	if (bcms_populate_signed_data(pool, signed_data,
				      content_type, data) != SECSuccess) {
		return NULL;
	}

	/* Add the signature. */
	status = bcms_add_signer_to_signed_data(pool,
						signed_data,
						digest_algorithm,
						SEC_OID_PKCS1_RSA_ENCRYPTION,
						preferred_slot,
						signer,
						pwcb_args);
	if (status != SECSuccess) {
		pkinit_debug("Error signing request!\n");
		return NULL;
	}

	/* Add any certs which we need to add beyond our own. */
	bcms_add_cert_chain_to_signed_data(pool, signed_data, certdb, signer);

	/* Encode the end result. */
	return bcms_encode_signed_data(pool, signed_data);
}

/* Make a copy of the list of recipients, ensuring that the passed-in value
 * is in there somewhere. */
static struct recipient_infos **
bcms_make_recipient_infos_list(PLArenaPool *pool, struct recipient_infos **list,
			       struct recipient_infos *recipient)
{
	PRBool found;
	struct recipient_infos **nlist;
	struct key_trans_recipient_info *a, *b;
	struct algorithm_identifier *alga, *algb;
	int i;

	found = PR_FALSE;
	for (i = 0; (list != NULL) && (list[i] != NULL) && !found; i++) {
		if (list[i]->recipient_info_type !=
		    recipient->recipient_info_type) {
			continue;
		}
		switch (list[i]->recipient_info_type) {
		case recipient_info_type_key_trans:
			a = &list[i]->recipient_info_value.ktri;
			b = &recipient->recipient_info_value.ktri;
			if (!SECITEM_ItemsAreEqual(&a->version, &b->version)) {
				continue;
			}
			if (a->rid.recipient_identifier_type !=
			    b->rid.recipient_identifier_type) {
				continue;
			}
			alga = &a->key_encryption_algorithm;
			algb = &b->key_encryption_algorithm;
			if (!SECITEM_ItemsAreEqual(&alga->algorithm,
						   &algb->algorithm)) {
				continue;
			}
			if (!SECITEM_ItemsAreEqual(alga->parameters,
						   algb->parameters)) {
				continue;
			}
			switch (a->rid.recipient_identifier_type) {
			case recipient_identifier_type_issuer_and_serial_number:
				if (!SECITEM_ItemsAreEqual(&a->rid.recipient_identifier_value.issuer_and_serial_number.issuer,
							   &b->rid.recipient_identifier_value.issuer_and_serial_number.issuer)) {
					continue;
				}
				if (!SECITEM_ItemsAreEqual(&a->rid.recipient_identifier_value.issuer_and_serial_number.serial,
							   &b->rid.recipient_identifier_value.issuer_and_serial_number.serial)) {
					continue;
				}
				break;
			case recipient_identifier_type_subject_key_identifier:
				if (!SECITEM_ItemsAreEqual(&a->rid.recipient_identifier_value.subject_key_identifier,
							   &b->rid.recipient_identifier_value.subject_key_identifier)) {
					continue;
				}
				break;
			default:
				break;
			}
			if (!SECITEM_ItemsAreEqual(&a->encrypted_key,
						   &b->encrypted_key)) {
				continue;
			}
			found = PR_TRUE;
			break;
		case recipient_info_type_key_agree:
			if (SECITEM_ItemsAreEqual(&list[i]->recipient_info_value.der_kari,
						  &recipient->recipient_info_value.der_kari)) {
				found = PR_TRUE;
				break;
			}
			break;
		case recipient_info_type_kek_recipient:
			if (SECITEM_ItemsAreEqual(&list[i]->recipient_info_value.der_kekri,
						  &recipient->recipient_info_value.der_kekri)) {
				found = PR_TRUE;
				break;
			}
			break;
		default:
			break;
		}
	}

	if (found) {
		nlist = list;
	} else {
		/* i contains the length of the current list */
		nlist = PORT_ArenaZAlloc(pool,
					 sizeof(struct recipient_infos *) *
					 (i + 2));
		if (nlist == NULL) {
			pkinit_debug("Out of memory\n");
			return NULL;
		}
		/* add our entry to the end of the list */
		nlist[i] = recipient;
		/* copy the rest of the list */
		while (--i >= 0) {
			nlist[i] = list[i];
		}
	}

	return nlist;
}

/* Find the certificate in the database which matches the recipient_identifier.
 */
static CERTCertificate *
find_cert_from_rid(CERTCertDBHandle *certdb, struct recipient_identifier *rid)
{
	struct issuer_and_serial_number *is;
	SECItem *spki;
	CERTIssuerAndSN isn;

	switch (rid->recipient_identifier_type) {
	case recipient_identifier_type_issuer_and_serial_number:
		/* The recipient is identified by issuer/serial. */
		is = &rid->recipient_identifier_value.issuer_and_serial_number;
		isn.derIssuer = is->issuer;
		isn.serialNumber = is->serial;
		return CERT_FindCertByIssuerAndSN(certdb, &isn);
		break;
	case recipient_identifier_type_subject_key_identifier:
		/* The recipient is identified by the key identifier. */
		spki = &rid->recipient_identifier_value.subject_key_identifier;
		return CERT_FindCertBySubjectKeyID(certdb, spki);
		break;
	default:
		break;
	}
	return NULL;
}

static SECItem *
bcms_recover_enveloped_data_key(PLArenaPool *pool,
				PK11SlotInfo *preferred_slot,
				CERTCertDBHandle *certdb,
				struct enveloped_data *env,
				struct pkinit_pwcb_args *pwcb_args)
{
	int i;
	struct recipient_infos *recipient_infos;
	struct key_trans_recipient_info *ktri;
	CERTCertificate *cert;
	SECKEYPrivateKey *privkey;
	SECOidData *oid;
	SECItem *key;
	SECStatus status;

	key = NULL;
	for (i = 0;
	     (env->recipient_infos != NULL) &&
	     (env->recipient_infos[i] != NULL) &&
	     (key == NULL);
	     i++) {
		recipient_infos = env->recipient_infos[i];
		/* we're looking for a transported key, nothing else will do */
		if (recipient_infos->recipient_info_type !=
		    recipient_info_type_key_trans) {
			pkinit_debug("recipient-info %d is not for key "
				     "transport!\n", i);
			continue;
		}
		/* we're looking for a transported key, nothing else will do */
		ktri = &recipient_infos->recipient_info_value.ktri;
		oid = SECOID_FindOID(&ktri->key_encryption_algorithm.algorithm);
		if (oid == NULL) {
			pkinit_debug("recipient-info %d uses an unrecognized "
				     "key encryption algorithm!\n", i);
			continue;
		} else {
			pkinit_debug("recipient-info %d uses \"%s\" to encrypt "
				     "the key\n", i, oid->desc);
		}
		/* ensure we know how to use that cipher */
		switch (oid->offset) {
		case SEC_OID_PKCS1_RSA_ENCRYPTION:
			break;
		default:
			pkinit_debug("recipient-info %d uses an unsupported "
				     "key encryption algorithm!\n", i);
			continue;
			break;
		}
		/* try to find a matching cert/key pair */
		cert = find_cert_from_rid(certdb, &ktri->rid);
		if (cert == NULL) {
			pkinit_debug("Couldn't find certificate for "
				     "recipient!\n");
			continue;
		}
		/* see if we have the matching private key */
		privkey = NULL;
		if (preferred_slot != NULL) {
			privkey = PK11_FindPrivateKeyFromCert(preferred_slot,
							      cert, pwcb_args);
		}
		if (privkey == NULL) {
			privkey = PK11_FindKeyByAnyCert(cert, pwcb_args);
		}
		if (privkey == NULL) {
			pkinit_debug("Unable to load private key.\n");
			continue;
		}
		/* try to decrypt the key data */
		switch (oid->offset) {
		case SEC_OID_PKCS1_RSA_ENCRYPTION:
			key = SECITEM_AllocItem(pool, NULL,
						ktri->encrypted_key.len);
			if (key != NULL) {
				status = PK11_PrivDecryptPKCS1(privkey,
							       key->data,
							       &key->len,
							       key->len,
							       ktri->encrypted_key.data,
							       ktri->encrypted_key.len);
				if (status != SECSuccess) {
					pkinit_debug("Error decrypting "
						     "envelope key!\n");
					key = NULL;
				} else {
					pkinit_debug("Decrypted enveloped "
						     "data key (%u bits)!\n",
						     key->len * 8);
				}
			} else {
				pkinit_debug("Out of memory!\n");
			}
			break;
		default:
			break;
		}
	}
	return key;
}

static SECItem *
bcms_add_padding(PLArenaPool *pool, SECItem *unpadded_data, int block_size)
{
	unsigned char *p;
	SECItem *ret;
	unsigned char pad_len;

	pad_len = (block_size - (unpadded_data->len % block_size));
	ret = SECITEM_AllocItem(pool, NULL, unpadded_data->len + pad_len);
	if (ret != NULL) {
		if (unpadded_data->len > 0) {
			memcpy(ret->data, unpadded_data->data,
			       unpadded_data->len);
		}
		p = ret->data + unpadded_data->len;
		while (p < ret->data + unpadded_data->len + pad_len) {
			*p = pad_len;
			p++;
		}
	}
	return ret;
}

static SECItem *
bcms_remove_padding(PLArenaPool *pool, SECItem *padded_data,
		    unsigned int block_size)
{
	unsigned char *p;
	unsigned char pad_len;
	SECItem *ret;

	if ((block_size > 1) && (padded_data->len >= block_size)) {
		p = padded_data->data + padded_data->len - 1;
		pad_len = *p;
		if (pad_len <= block_size) {
			ret = SECITEM_AllocItem(pool, NULL,
						padded_data->len - pad_len);
			if (ret != NULL) {
				memcpy(ret->data, padded_data->data,
				       padded_data->len - pad_len);
				return ret;
			}
		}
	}
	return NULL;
}

/* Recover the data from an enveloped_data structure. */
SECItem *
bcms_extract_enveloped_data(PLArenaPool *pool,
			    CERTCertDBHandle *certdb,
			    struct enveloped_data *env,
			    PK11SlotInfo *preferred_slot,
			    struct pkinit_pwcb_args *pwcb_args)
{
	SECOidData *oid;
	SECItem *key;
	struct encrypted_content_info *eci;
	struct algorithm_identifier *enc_alg;
	struct rc2_cbc_parameters rc2_params;
	unsigned long rc2_version;
	SECItem des3_params;
	CK_RC2_CBC_PARAMS rc2_cbc_params;
	CK_MECHANISM mech;
	PK11SlotInfo *slot;
	PK11SymKey *symkey;
	PK11Context *ctx;
	SECItem params, *plaintext;
	int outlen;
	unsigned int ublock_size;
	int block_size;

	eci = &env->encrypted_content_info;
	enc_alg = &eci->content_encryption_algorithm;

	/* Allocate the decryption buffer. */
	plaintext = SECITEM_AllocItem(pool, NULL, eci->encrypted_content.len);
	if (plaintext == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}

	/* Recover the payload encryption key. */
	key = bcms_recover_enveloped_data_key(pool, preferred_slot, certdb,
					      env, pwcb_args);
	if (key == NULL) {
		pkinit_debug("Error recovering temporary key!\n");
		return NULL;
	}

	/* Spot-check that we'll be able to decrypt the payload, and then parse
	 * out the algorithm-specific parameters. */
	oid = SECOID_FindOID(&enc_alg->algorithm);
	if (oid == NULL) {
		pkinit_debug("Unrecognized encryption algorithm!\n");
		return NULL;
	}
	memset(&mech, 0, sizeof(mech));
	switch (oid->offset) {
	case SEC_OID_RC2_CBC:
		block_size = 8;
		ublock_size = 8;
		if (enc_alg->parameters == NULL) {
			pkinit_debug("Expected RC2-CBC parameters, "
				     "got none!\n");
			return NULL;
		}
		if (SEC_ASN1DecodeItem(pool, &rc2_params,
				       common_rc2_cbc_parameters_template,
				       enc_alg->parameters) != SECSuccess) {
			pkinit_debug("Error parsing RC2-CBC parameters!\n");
			return NULL;
		}
		if (SEC_ASN1DecodeInteger(&rc2_params.version,
					  &rc2_version) != SECSuccess) {
			pkinit_debug("Error decoding RC2-CBC \"version\"!\n");
			return NULL;
		}
		if (rc2_params.iv.len != ublock_size) {
			pkinit_debug("RC2-CBC IV is of unexpected size "
				     "(expected %d bits, got %d)!\n",
				     ublock_size * 8, rc2_params.iv.len * 8);
			return NULL;
		}

		rc2_cbc_params.ulEffectiveBits = rc2_version;
		/* FIXME: is the effective key size ever different from the
		 * real one?  Do we really have to map this "version" to an
		 * effective key size? */
		rc2_cbc_params.ulEffectiveBits = key->len * 8;
		memcpy(&rc2_cbc_params.iv[0], rc2_params.iv.data, block_size);
		mech.mechanism = CKM_RC2_CBC;
		mech.pParameter = &rc2_cbc_params;
		mech.ulParameterLen = sizeof(&rc2_cbc_params);
		break;
	case SEC_OID_DES_EDE3_CBC:
		block_size = 8;
		ublock_size = 8;
		if (enc_alg->parameters == NULL) {
			pkinit_debug("Expected DES3 parameters, got none!\n");
			return NULL;
		}
		if (SEC_ASN1DecodeItem(pool, &des3_params,
				       SEC_OctetStringTemplate,
				       enc_alg->parameters) != SECSuccess) {
			pkinit_debug("Error parsing DES3 parameters!\n");
			return NULL;
		}
		if (des3_params.len != ublock_size) {
			pkinit_debug("RC2-CBC IV is of unexpected size "
				     "(expected %d bits, got %d)!\n",
				     ublock_size * 8, des3_params.len * 8);
			return NULL;
		}
		mech.mechanism = CKM_DES3_CBC;
		mech.pParameter = des3_params.data;
		mech.ulParameterLen = des3_params.len;
		break;
	default:
		pkinit_debug("Unsupported encryption algorithm!\n");
		return NULL;
	}

	/* Use the PKCS11 interface to decrypt the data. */
	slot = PK11_GetBestSlot(mech.mechanism, pwcb_args);
	if (slot == NULL) {
		pkinit_debug("Can't find slot!\n");
		return NULL;
	}
	symkey = PK11_ImportSymKey(slot, mech.mechanism, PK11_OriginUnwrap,
				   CKA_DECRYPT, key, pwcb_args);
	if (symkey == NULL) {
		pkinit_debug("Can't import temporary key!\n");
		return NULL;
	}
	params.data = mech.pParameter;
	params.len = mech.ulParameterLen;
	ctx = PK11_CreateContextBySymKey(mech.mechanism, CKA_DECRYPT,
					 symkey, &params);
	if (ctx == NULL) {
		pkinit_debug("Can't initialize decryption context!\n");
		return NULL;
	}

	if ((PK11_CipherOp(ctx, plaintext->data, &outlen, plaintext->len,
			   eci->encrypted_content.data,
			   eci->encrypted_content.len) != SECSuccess) ||
	    (PK11_Finalize(ctx) != SECSuccess)) {
		pkinit_debug("Error decrypting encapsulated content!\n");
		return NULL;
	}
	plaintext->len = outlen;

	/* Make a copy of the plaintext to the caller. */
	return bcms_remove_padding(pool, plaintext, ublock_size);
}

/* Add a new recipient_infos for the passed-in certificate to the list of
 * recipients in the enveloped-data structure. */
static SECStatus
bcms_add_recipient_to_enveloped_data(PLArenaPool *pool,
				     struct enveloped_data *env,
				     CERTCertificate *recipient,
				     SECItem *content_enc_key)
{
	struct recipient_infos *rec;
	struct key_trans_recipient_info *trans;
	struct issuer_and_serial_number *isn;
	SECItem *enc;
	SECOidData *oid;
	SECKEYPublicKey *key;

	rec = PORT_ArenaZAlloc(pool, sizeof(*rec));
	if (rec == NULL) {
		pkinit_debug("Out of memory.\n");
		return SECFailure;
	}

	/* Mark that this is a key-transport recipient-infos. */
	rec->recipient_info_type = recipient_info_type_key_trans;

	/* Set the recipient version (1), indicating that it's issuer-serial
	 * identification. */
	trans = &rec->recipient_info_value.ktri;
	if (SEC_ASN1EncodeInteger(pool, &trans->version, 1) !=
	    &trans->version) {
		pkinit_debug("Internal error.\n");
		return SECFailure;
	}

	/* Set the recipient's issuer-serial pair. */
	trans->rid.recipient_identifier_type =
		recipient_identifier_type_issuer_and_serial_number;
	isn = &trans->rid.recipient_identifier_value.issuer_and_serial_number;
	isn->issuer = recipient->derIssuer;
	isn->serial = recipient->serialNumber;

	/* Recover the recipient's public key. */
	key = CERT_ExtractPublicKey(recipient);
	if (key == NULL) {
		pkinit_debug("Internal error.\n");
		return SECFailure;
	}
	if (key->keyType != rsaKey) {
		pkinit_debug("Unsupported client public key type.\n");
		return SECFailure;
	}

	/* Describe how we'll be encrypting the content-encryption key. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS1_RSA_ENCRYPTION);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return SECFailure;
	}
	trans->key_encryption_algorithm.algorithm = oid->oid;
	trans->key_encryption_algorithm.parameters = &bcms_null_item;

	/* Encrypt the content-encryption key. */
	pkinit_debug("Encrypting %d bytes of CEK data for \"%s\".\n",
		     content_enc_key->len, recipient->nickname);
	enc = SECITEM_AllocItem(pool, NULL, key->u.rsa.modulus.len);
	if (PK11_PubEncryptPKCS1(key, enc->data,
				 content_enc_key->data, content_enc_key->len,
				 NULL) != SECSuccess) {
		pkinit_debug("Error encrypting content-encryption key.\n");
		return SECFailure;
	}
	trans->encrypted_key = *enc;

	/* Finally, add the recipient info to the list. */
	env->recipient_infos = bcms_make_recipient_infos_list(pool,
							      env->recipient_infos,
							      rec);
	return SECSuccess;
}

static SECStatus
bcms_populate_enveloped_data(PLArenaPool *pool,
			     struct enveloped_data *enveloped_data,
			     const SECItem *encrypted_content_type)
{
	struct encrypted_content_info *encrypted_info;
	SECOidData *oid;

	/* Set the version number. */
	if (SEC_ASN1EncodeInteger(pool, &enveloped_data->version, 0) !=
	    &enveloped_data->version) {
		pkinit_debug("Internal error.\n");
		return SECFailure;
	}

	/* We have no sender identification. */
	memset(&enveloped_data->der_originator_info, 0,
	       sizeof(enveloped_data->der_originator_info));

	/* Initialize as empty the list of recipients. */
	enveloped_data->recipient_infos = NULL;

	/* We have no unprotected attributes. */
	memset(&enveloped_data->der_unsigned_attrs, 0,
	       sizeof(enveloped_data->der_unsigned_attrs));

	/* Fill in the content type of the encrypted content-info. */
	encrypted_info = &enveloped_data->encrypted_content_info;
	if (encrypted_content_type == NULL) {
		/* Retrieve the OID for the default of "data". */
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
		if (oid == NULL) {
			return SECFailure;
		}
		encrypted_content_type = &oid->oid;
	}
	encrypted_info->content_type = *encrypted_content_type;

	return SECSuccess;
}

static SECStatus
bcms_encrypt_with_key(CK_MECHANISM *mech,
		      PK11Origin key_origin,
		      SECItem *key,
		      SECItem *padded_plaintext,
		      SECItem *ciphertext)
{
	PK11SlotInfo *slot;
	PK11Context *ctx;
	PK11SymKey *symkey;
	SECItem params;
	int outlen;

	/* Actually encrypt the payload, using the content encryption key. */
	slot = PK11_GetBestSlot(mech->mechanism, NULL);
	if (slot == NULL) {
		pkinit_debug("Can't find slot!\n");
		return SECFailure;
	}
	symkey = PK11_ImportSymKey(slot, mech->mechanism, key_origin,
				   CKA_ENCRYPT, key, NULL);
	if (symkey == NULL) {
		pkinit_debug("Can't import content encryption key!\n");
		return SECFailure;
	}
	params.data = mech->pParameter;
	params.len = mech->ulParameterLen;
	ctx = PK11_CreateContextBySymKey(mech->mechanism, CKA_ENCRYPT,
					 symkey, &params);
	if (ctx == NULL) {
		pkinit_debug("Can't initialize encryption context!\n");
		return SECFailure;
	}
	if ((PK11_CipherOp(ctx,
			   ciphertext->data,
			   &outlen,
			   ciphertext->len,
			   padded_plaintext->data,
			   padded_plaintext->len) != SECSuccess) ||
	    (PK11_Finalize(ctx) != SECSuccess)) {
		pkinit_debug("Error encrypting encapsulated content!\n");
		return SECFailure;
	}
	ciphertext->len = outlen;
	return SECSuccess;
}

/* Build a new encoded enveloped_data. */
SECItem *
bcms_create_enveloped_data(PLArenaPool *pool,
			   CERTCertDBHandle *certdb,
			   krb5_context kcontext,
			   SECItem *content_type,
			   SECOidTag preferred_cek_algorithm,
			   SECItem *data,
			   SECItem *recipient_cert)
{
	struct enveloped_data *enveloped_data;
	struct encrypted_content_info *encrypted_info;
	struct algorithm_identifier *algorithm;
	CERTCertificate **recipients;
	SECItem content_enc_key, *ciphertext, *plaintext;
	CK_MECHANISM mech;
	int block_size;
	unsigned int ublock_size;

	pkinit_debug("Generating enveloped-data.\n");

	/* Create and populate the enveloped data structure. */
	enveloped_data = PORT_ArenaZAlloc(pool, sizeof(*enveloped_data));
	if (enveloped_data == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}

	/* Build the enveloped_data structure. */
	if (SEC_ASN1EncodeInteger(pool, &enveloped_data->version, 0) !=
	    &enveloped_data->version) {
		pkinit_debug("Internal error.\n");
		return NULL;
	}

	if (bcms_populate_enveloped_data(pool, enveloped_data,
					 content_type) != SECSuccess) {
		return NULL;
	}

	/* Note the content encryption algorithm. */
	encrypted_info = &enveloped_data->encrypted_content_info;
	algorithm = &encrypted_info->content_encryption_algorithm;
	if (common_generate_content_encryption_key(pool,
						   kcontext,
						   preferred_cek_algorithm,
						   algorithm,
						   &content_enc_key,
						   &mech,
						   &block_size)) {
		pkinit_debug("Error generating content encryption key!\n");
		return NULL;
	}
	ublock_size = block_size;

	/* Pad out the data to get our plaintext. */
	plaintext = bcms_add_padding(pool, data, ublock_size);
	if (plaintext == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}

	/* Allocate room for the destination buffer.  We're just guessing that
	 * this is the right size. */
	ciphertext = SECITEM_AllocItem(pool, NULL,
				       plaintext->len + (block_size * 2));
	if (ciphertext == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}

	/* Actually encrypt the payload, using the content encryption key. */
	pkinit_debug("Encrypting %d bytes of data (%d once padded) with CEK.\n",
		     data->len, plaintext->len);
	if (bcms_encrypt_with_key(&mech,
				  PK11_OriginGenerated,
				  &content_enc_key,
				  plaintext,
				  ciphertext) != SECSuccess) {
		pkinit_debug("Error encrypting encapsulated content!\n");
		return NULL;
	}
	encrypted_info->encrypted_content = *ciphertext;

	/* Wrap up a copy of the CEK for the client. */
	recipients = NULL;
	if (CERT_ImportCerts(certdb, certUsageSSLClient,
			     1,
			     &recipient_cert,
			     &recipients,
			     PR_FALSE,
			     PR_FALSE,
			     NULL) != SECSuccess) {
		pkinit_debug("Error parsing client certificate!\n");
		return NULL;
	}
	if (bcms_add_recipient_to_enveloped_data(pool,
						 enveloped_data,
						 recipients[0],
						 &content_enc_key) != SECSuccess) {
		pkinit_debug("Error encrypting CEK for recipient!\n");
		return NULL;
	}

	/* Encode the end result. */
	return bcms_encode_enveloped_data(pool, enveloped_data);
}
