/*
 * Copyright (C) 2006 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: certs.c,v 1.13 2006/11/01 16:32:59 nalin Exp $"

#include "../config.h"

#include <stdio.h>

#include <krb5/krb5.h>

#include <nss.h>
#include <cert.h>
#include <certdb.h>
#include <secmodt.h>

#include "commont.h"
#include "bcmst.h"
#include "pkinitt.h"
#include "certs.h"
#include "pkinit.h"

#ifdef USE_PAL_BACKPORT
#include "backport-errors.h"
#endif

static unsigned char oid_ms_sc_login_data[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x14, 0x02, 0x02};
static const SECOidData
oid_ms_sc_login = {
	.oid = {
		.data = oid_ms_sc_login_data,
		.len = 10,
	},
	.offset = 0,
	.desc = "Microsoft Windows SmartCard Logon",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_ms_upn_name_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x14, 0x02, 0x03};
static const SECOidData
oid_ms_upn_name = {
	.oid = {
		.data = oid_ms_upn_name_bytes,
		.len = 10,
	},
	.offset = 0,
	.desc = "Microsoft Windows User Principal Name",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char
oid_ms_guid_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x19, 0x01};
static const SECOidData
oid_ms_guid = {
	.oid = {
		.data = oid_ms_guid_bytes,
		.len = 9,
	},
	.offset = 0,
	.desc = "Microsoft Globally-Unique Identifier",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char
oid_ms_certificate_template_name_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x14, 0x02};
static const SECOidData
oid_ms_certificate_template_name = {
	.oid = {
		.data = oid_ms_certificate_template_name_bytes,
		.len = 9,
	},
	.offset = 0,
	.desc = "Microsoft Certificate Template Name",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

/* pkinit-SAN 1.3.6.1.5.2.2 */
static unsigned char oid_pkinit_san_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x02};
static const SECOidData
oid_pkinit_san = {
	.oid = {
		.data = oid_pkinit_san_bytes,
		.len = 6,
	},
	.offset = 0,
	.desc = "PKINIT Subject Alternate Name",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_auth_data_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x01};
static const SECOidData
oid_pkinit_auth_data = {
	.oid = {
		.data = oid_pkinit_auth_data_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Authentication Data",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_rkey_data_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x03};
static const SECOidData
oid_pkinit_rkey_data = {
	.oid = {
		.data = oid_pkinit_rkey_data_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Reply Key Data",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_dhkey_data_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x02};
static const SECOidData
oid_pkinit_dhkey_data = {
	.oid = {
		.data = oid_pkinit_dhkey_data_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Diffie-Hellman Reply Key Data",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_key_purpose_client_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x04};
static const SECOidData
oid_pkinit_key_purpose_client = {
	.oid = {
		.data = oid_pkinit_key_purpose_client_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Key Purpose: Client",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_key_purpose_kdc_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x05};
static const SECOidData
oid_pkinit_key_purpose_kdc = {
	.oid = {
		.data = oid_pkinit_key_purpose_kdc_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Key Purpose: KDC",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_dh_public_number_bytes[] = {0x2a, 0x86, 0x48, 0xce, 0x3e, 0x02, 0x01};
static const SECOidData
oid_dh_public_number = {
	.oid = {
		.data = oid_dh_public_number_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "Diffie-Hellman Public Number",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_dh_key_agreement_bytes[] = {0x2a, 0x86, 0x48, 0x86, 0xf6, 0x33, 0x01, 0x03, 0x01};
static const SECOidData
oid_dh_key_agreement = {
	.oid = {
		.data = oid_dh_key_agreement_bytes,
		.len = 9,
	},
	.offset = 0,
	.desc = "Diffie-Hellman Key Agreement",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

/* Template for an otherName.  If it's a MS-UPN, then the data is an explicit
 * UTF8String. If it's a PKINIT-SAN, it's a structure. */
static SEC_ASN1Template
pkinit_name_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = 0,
	.sub = pkinit_kerberos_name_template,
	.size = sizeof(struct kerberos_name),
	},
};
static SEC_ASN1Template
ms_upn_name_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = 0,
	.sub = SEC_UTF8StringTemplate,
	.size = sizeof(SECItem),
	},
};
static SEC_ASN1Template
ms_guid_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = 0,
	.sub = SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
};
static SEC_ASN1Template
other_name_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct other_name),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct other_name, oid),
	.sub = NULL,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED,
	.offset = offsetof(struct other_name, data),
	.sub = SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

/* Template for decoding the choice of subjectAltName. */
static SEC_ASN1Template
san_choice_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_type), /* id */
	.sub = NULL,
	.size = sizeof(struct subject_alt_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_value.other_name),
	.sub = other_name_template,
	.size = subject_alt_name_other_name, /* identifier to save for this choice */
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_value.rfc822_name),
	.sub = &SEC_IA5StringTemplate,
	.size = subject_alt_name_rfc822_name, /* identifier to save for this choice */
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_value.dnsname),
	.sub = &SEC_IA5StringTemplate,
	.size = subject_alt_name_dnsname, /* identifier to save for this choice */
	},
	{0, 0, NULL, 0},
};

/* Template for unpacking the subjectAltName data sequence into an
 * array of other_name structures. */
static SEC_ASN1Template
san_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = &san_choice_template,
	.size = 0,
	},
};

/* Template for unpacking an EKU list. */
static SEC_ASN1Template
eku_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = SEC_ObjectIDTemplate,
	.size = 0,
	},
};

const SECItem *
cert_get_oid_pkinit_auth_data(void)
{
	return &oid_pkinit_auth_data.oid;
}
const SECItem *
cert_get_oid_pkinit_rkey_data(void)
{
	return &oid_pkinit_rkey_data.oid;
}
const SECItem *
cert_get_oid_pkinit_dhkey_data(void)
{
	return &oid_pkinit_dhkey_data.oid;
}
const SECItem *
cert_get_oid_dh_public_number(void)
{
	return &oid_dh_public_number.oid;
}
const SECItem *
cert_get_oid_dh_key_agreement(void)
{
	return &oid_dh_key_agreement.oid;
}

/* Check if the certificate contains the desired OID as an EKU value. */
static SECStatus
cert_eku_matches_oid(CERTCertificate *cert, const SECItem *oid, PRBool *matches)
{
	SECItem eku_value;
	const SECItem **oids;
	SECOidData *eku;
	int i;

	eku = SECOID_FindOIDByTag(SEC_OID_X509_EXT_KEY_USAGE);
	if (eku == NULL) {
		pkinit_debug("Internal error!\n");
		return SECFailure;
	}
	if (CERT_FindCertExtension(cert, eku->offset,
				   &eku_value) == SECSuccess) {
		oids = NULL;
		*matches = PR_FALSE;
		if (SEC_ASN1DecodeItem(cert->arena, &oids, eku_template,
				       &eku_value) == SECSuccess) {
			for (i = 0; (oids != NULL) && (oids[i] != NULL); i++) {
				if (SECITEM_ItemsAreEqual(oids[i], oid)) {
					pkinit_debug("Found desired EKU.\n");
					*matches = PR_TRUE;
					break;
				}
			}
			return SECSuccess;
		} else {
			pkinit_debug("Error parsing extendedKeyUsage "
				     "extension!\n");
			return SECFailure;
		}
	}

	pkinit_debug("Certificate has no extendedKeyUsage extension!\n");
	return SECFailure;
}

static void
derdump(SECItem *item)
{
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		fprintf(stderr, "Dumping %d bytes.\n", item->len);
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
}

/* Check that the keyUsage extension includes the specified bits. */
static SECStatus
cert_ku_matches_mask(CERTCertificate *cert, unsigned int ku_mask,
		     PRBool *matches)
{
	*matches = cert->keyUsage & ku_mask;
	return cert->keyUsagePresent ? SECSuccess : SECFailure;
}

/* Check if the certificate contains the tagged OID as an EKU value. */
static SECStatus
cert_eku_matches_tag(CERTCertificate *cert, SECOidTag eku, PRBool *matches)
{
	SECOidData *oid;
	oid = SECOID_FindOIDByTag(eku);
	if (oid == NULL) {
		pkinit_debug("No EKU extension found!\n");
		return SECFailure;
	}
	return cert_eku_matches_oid(cert, &oid->oid, matches);
}

/* Compare a SECItem and a krb5_data item. */
static PRBool
cert_secitem_equals_krb5_data(SECItem *item, krb5_data *data)
{
	return (item->len == data->length) &&
	       (memcmp(item->data, data->data, item->len) == 0);
}
static PRBool
cert_secitem_list_equals_krb5_data_list(SECItem **items, krb5_data *list,
					int length)
{
	int i;
	for (i = 0; i < length; i++) {
		if ((items == NULL) || (items[i] == NULL)) {
			return PR_FALSE;
		}
	}
	if (items[i] != NULL) {
		return PR_FALSE;
	}
	for (i = 0; i < length; i++) {
		if (!cert_secitem_equals_krb5_data(items[i], &list[i])) {
			return PR_FALSE;
		}
	}
	return PR_TRUE;
}

/* Retrieve an extension based on its OID. */
static SECStatus
cert_find_extension_oid(CERTCertificate *cert,
			const SECItem *oid, SECItem *value)
{
	int i;
	for (i = 0;
	     (cert->extensions != NULL) && (cert->extensions[i] != NULL);
	     i++) {
		if (SECITEM_ItemsAreEqual(&cert->extensions[i]->id, oid)) {
			if (value != NULL) {
				*value = cert->extensions[i]->value;
			}
			return SECSuccess;
		}
	}
	return SECFailure;
}

/* Check if the certificate certificateTemplateName matches the given string. */
static SECStatus
cert_template_name_matches(CERTCertificate *cert, const char *ascii_value,
			   PRBool *matches)
{
	SECItem template_name, name;
	unsigned int i;
	if (cert_find_extension_oid(cert,
				    &oid_ms_certificate_template_name.oid,
				    &template_name) != SECSuccess) {
		pkinit_debug("No certificateTemplateName extension!\n");
		return SECFailure;
	}
	if (SEC_ASN1DecodeItem(cert->arena, &name,
			       SEC_BMPStringTemplate,
			       &template_name) != SECSuccess) {
		pkinit_debug("Error decoding certificateTemplateName!\n");
		return SECFailure;
	}
	/* Yes, this is cheap. But short of doing character set conversion,
	 * it'll do for now. */
	for (i = 0; ascii_value[i] != '\0'; i++) {
		/* Are we out of certificateTemplateName value? */
		if ((i * 2) >= name.len) {
			pkinit_debug("certificateTemplateName was shorter than "
				     "we expected!\n");
			break;
		}
		/* Match one character. */
		if ((((char*)name.data)[i * 2] != '\0') ||
		    (((char*)name.data)[i * 2 + 1] != ascii_value[i])) {
			break;
		}
	}
	if ((i > 0) && (ascii_value[i] == '\0') && ((i * 2) == name.len)) {
		*matches = PR_TRUE;
		pkinit_debug("certificateTemplateName matched \"%s\"\n",
			     ascii_value);
	} else {
		pkinit_debug("certificateTemplateName did not match \"%s\"\n",
			     ascii_value);
		pkinit_debug("certificateTemplateName value was \"%.*s\"\n",
			     name.len, name.data);
	}
	return SECSuccess;
}

/* Check if the certificate subjectAltName UPN value matches the principal. */
static SECStatus
cert_san_matches_upn(CERTCertificate *cert,
		     krb5_context kcontext, krb5_principal principal,
		     PRBool *matches)
{
	struct subject_alt_name **names;
	SECItem san_value, unparsed_name, ms_upn_name;
	char *unparsed;
	int i;

	i = 0;
	names = NULL;

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug("No subjectAltName extension in certificate!\n");
		return SECFailure;
	}

	/* Split up the subjectAltName sequence. */
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug("Error parsing subjectAltName extension!\n");
		return SECFailure;
	}

	/* Create the UTF8String representation of the principal name. */
	if (krb5_unparse_name(kcontext, principal, &unparsed) != 0) {
		pkinit_debug("Error unparsing krb5_principal!\n");
		return SECFailure;
	}
	unparsed_name.data = (unsigned char *) unparsed;
	unparsed_name.len = strlen(unparsed);

	/* Iterate over all of the values. */
	*matches = PR_FALSE;
	for (i = 0; (names != NULL) && (names[i] != NULL) && !(*matches); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_other_name:
			/* Check for a Microsoft UPN value. */
			if (SECITEM_ItemsAreEqual(&names[i]->subject_alt_name_value.other_name.oid,
						  &oid_ms_upn_name.oid)) {
				pkinit_debug("Checking UPN.\n");
				/* If we can decode the UTF-8 string. */
				if (SEC_ASN1DecodeItem(cert->arena,
						       &ms_upn_name,
						       ms_upn_name_template,
						       &names[i]->subject_alt_name_value.other_name.data) == SECSuccess) {
					/* And it matches, then we're okay. */
					if (SECITEM_ItemsAreEqual(&ms_upn_name,
								  &unparsed_name)) {
						pkinit_debug("UPN Matched!\n");
						*matches = PR_TRUE;
					} else {
						pkinit_debug("\"%.*s\" != "
							     "\"%.*s\"\n",
							     ms_upn_name.len,
							     ms_upn_name.data,
							     unparsed_name.len,
							     unparsed_name.data);
					}
				} else {
					pkinit_debug("Error decoding UPN!\n");
				}
			} else {
				pkinit_debug("SAN %d is not a UPN.\n", i + 1);
			}
			break;
		default:
			pkinit_debug("Name %d, type %d, is not an otherName.\n",
				     i + 1, names[i]->subject_alt_name_type);
			break;
		};
	}
	krb5_free_unparsed_name(kcontext, unparsed);
	pkinit_debug("%d names found\n", i);
	return SECSuccess;
}

/* Check if the certificate subjectAltName matches the principal. */
static SECStatus
cert_san_matches_principal(CERTCertificate *cert,
			   krb5_principal principal,
			   PRBool *matches)
{
	struct subject_alt_name **names;
	SECItem san_value;
	struct kerberos_name kerberos_name;
	int i;

	i = 0;
	names = NULL;

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug("No subjectAltName extension in certificate!\n");
		return SECFailure;
	}

	/* Split up the subjectAltName sequence. */
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug("Error parsing subjectAltName extension!\n");
		return SECFailure;
	}

	/* Iterate over all of the values. */
	*matches = PR_FALSE;
	for (i = 0; (names != NULL) && (names[i] != NULL) && !(*matches); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_other_name:
			/* Check for a PKINIT SAN value. */
			if (SECITEM_ItemsAreEqual(&names[i]->subject_alt_name_value.other_name.oid,
						  &oid_pkinit_san.oid)) {
				pkinit_debug("Checking Kerberos name.\n");
				/* If we can decode the PKINIT SAN. */
				if (SEC_ASN1DecodeItem(cert->arena,
						       &kerberos_name,
						       pkinit_name_template,
						       &names[i]->subject_alt_name_value.other_name.data) == SECSuccess) {
					/* And it matches, then we're okay. */
					if (cert_secitem_equals_krb5_data(&kerberos_name.realm.name, krb5_princ_realm(kcontext, principal)) &&
					    cert_secitem_list_equals_krb5_data_list(kerberos_name.principal_name.name_string, principal->data, krb5_princ_size(kcontext, principal))) {
						pkinit_debug("Principal matched!\n");
						*matches = PR_TRUE;
					}
				} else {
					pkinit_debug("Error decoding principal "
						     "name!\n");
				}
			} else {
				pkinit_debug("SAN %d is not a Kerberos name.\n",
					     i + 1);
			}
			break;
		default:
			pkinit_debug("Name %d, type %d, is not an otherName.\n",
				     i + 1, names[i]->subject_alt_name_type);
			break;
		};
	}
	pkinit_debug("%d names found\n", i);
	return SECSuccess;
}

PRBool
cert_is_preferred(CERTCertificate *cert,
		  krb5_context kcontext, krb5_principal principal,
		  PRInt32 cert_flags)
{
	PRBool match;
	if (cert == NULL) {
		return PR_FALSE;
	}
	if ((cert_flags & CERT_SAN_MASK) == 0) {
		return PR_FALSE;
	}
	/* The PKINIT certificate will have our principal name encoded as a
	 * krb5PrincipalName. */
	if (cert_flags & CERT_MATCHES_KRB_SAN) {
		cert_flags &= ~CERT_MATCHES_KRB_SAN;
		if ((cert_san_matches_principal(cert,
						principal,
						&match) != SECSuccess) ||
		    !match) {
			if (!(cert_flags & CERT_SAN_MASK)) {
				/* Return the error if we have no more name
				 * types to try matching. */
				return PR_FALSE;
			}
		}
	}
	/* The Windows certificate will have our principal name encoded as a
	 * userPrincipalName, and *may* have the Client Authentication and Smart
	 * Card Logon EKUs set. */
	if (cert_flags & CERT_MATCHES_UPN_SAN) {
		cert_flags &= ~CERT_MATCHES_UPN_SAN;
		if ((cert_san_matches_upn(cert,
					  kcontext, principal,
					  &match) != SECSuccess) ||
		    !match) {
			if (!(cert_flags & CERT_SAN_MASK)) {
				/* Return the error if we have no more name
				 * types to try matching. */
				return PR_FALSE;
			}
		}
	}
	/* Now check for the EKU values which were specified: normal client
	 * authentication. */
	if (cert_flags & CERT_MATCHES_CLIENT_AUTH_EKU) {
		if ((cert_eku_matches_tag(cert,
					  SEC_OID_EXT_KEY_USAGE_CLIENT_AUTH,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* Microsoft Smartcard Logon */
	if (cert_flags & CERT_MATCHES_MS_SMARTCARD_EKU) {
		if ((cert_eku_matches_oid(cert,
					  &oid_ms_sc_login.oid,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* Server Authentication */
	if (cert_flags & CERT_MATCHES_SERVER_AUTH_EKU) {
		if ((cert_eku_matches_tag(cert,
					  SEC_OID_EXT_KEY_USAGE_SERVER_AUTH,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* PKINIT Client */
	if (cert_flags & CERT_MATCHES_PKINIT_CLIENT_EKU) {
		if ((cert_eku_matches_oid(cert,
					  &oid_pkinit_key_purpose_client.oid,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* PKINIT KDC */
	if (cert_flags & CERT_MATCHES_PKINIT_KDC_EKU) {
		if ((cert_eku_matches_oid(cert,
					  &oid_pkinit_key_purpose_kdc.oid,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	return PR_TRUE;
}

CERTCertificate *
cert_find_preferred_cert_slot(PK11SlotInfo *slot,
			      struct pkinit_pwcb_args *pwcb_args,
			      krb5_context kcontext,
			      krb5_principal principal,
			      PRInt32 cert_flags)
{
	CERTCertificate *cert, *ret = NULL;
	CERTCertList *cert_list;
	CERTCertListNode *node;

	pkinit_debug("Checking slot \"%s\" for certificates\n",
		     PK11_GetSlotName(slot));

	if (PK11_Authenticate(slot, PR_TRUE, pwcb_args) != SECSuccess) {
		pkinit_debug("PK11_Authenticate failed for slot %s\n",
			     PK11_GetSlotName(slot));
		return NULL;
	}

	cert_list = PK11_ListCertsInSlot(slot);
	if ((cert_list == NULL) || CERT_LIST_EMPTY(cert_list)) {
		return NULL;
	}

	for (node = CERT_LIST_HEAD(cert_list);
	     !CERT_LIST_END(node, cert_list);
	     node = CERT_LIST_NEXT(node)) {

		cert = node->cert;
		if (cert == NULL) {
			continue;
		}

		pkinit_debug("Checking \"%s\".\n", cert->nickname);
		if (cert_is_preferred(cert, kcontext,
				      principal, cert_flags)) {
			pkinit_debug("Using \"%s\"\n", cert->nickname);
			/* CERT_DestroyCertList will free 'cert', so we need
			 * to make a copy of it. */
			ret = CERT_DupCertificate(cert);
			if (!ret) {
				pkinit_debug("%s: failed to allocate memory "
					     "to hold a certificate\n",
					     __FUNCTION__);
				break;
			}
			break;
		}
	}

	CERT_DestroyCertList(cert_list);
	return ret;
}

/* Search, first the preferred slot, and then everywhere, for a certificate
 * which matches the named principal. */
CERTCertificate *
cert_find_preferred_cert(PK11SlotInfo *preferred_slot,
			 struct pkinit_pwcb_args *pwcb_args,
			 krb5_context kcontext,
			 krb5_principal principal,
			 PRInt32 cert_flags)
{
	CERTCertificate *cert;
	PK11SlotList *slot_list;
	PK11SlotListElement *le;

	if (preferred_slot) {
		cert = cert_find_preferred_cert_slot(preferred_slot, pwcb_args,
						     kcontext, principal,
						     cert_flags);
		if (cert) {
			return cert;
		}
	}

	/* Failing that, try them all. */
	slot_list = PK11_GetAllTokens(CKM_INVALID_MECHANISM,
				      PR_FALSE, PR_FALSE, NULL);
	if (slot_list == NULL) {
		pkinit_debug("PK11_GetAllTokens returned no results!\n");
		return NULL;
	}

	/* first pass, only try those slots that are already logged-in */
	for (le = slot_list->head; le != NULL; le = le->next) {
		if (PK11_NeedLogin(le->slot)) {
			continue;
		}

		cert = cert_find_preferred_cert_slot(le->slot, pwcb_args,
						     kcontext, principal,
						     cert_flags);
		if (cert) {
			PK11_FreeSlotList(slot_list);
			return cert;
		}
	}

	/* Second pass: check certificates on slots which require login */
	for (le = slot_list->head; le != NULL; le = le->next) {
		if (!PK11_NeedLogin(le->slot)) {
			continue;
		}

		cert = cert_find_preferred_cert_slot(le->slot, pwcb_args,
						     kcontext, principal,
						     cert_flags);
		if (cert != NULL) {
			PK11_FreeSlotList(slot_list);
			return cert;
		}
	}

	PK11_FreeSlotList(slot_list);
	return NULL;
}

/* Try to determine if the certificate is a "CA" certificate. */
static SECStatus
cert_certificate_get_is_ca(CERTCertificate *cert, PRBool *is_ca)
{
	CERTBasicConstraints basic_constraints;
	SECItem value;

	if (CERT_FindCertExtension(cert, SEC_OID_X509_BASIC_CONSTRAINTS,
				   &value) != SECSuccess) {
		pkinit_debug("Error retrieving basic constraints value!\n");
		return SECFailure;
	}
	if (CERT_DecodeBasicConstraintValue(&basic_constraints,
					    &value) != SECSuccess) {
		pkinit_debug("Error decoding basic constraints value!\n");
		return SECFailure;
	}
	*is_ca = basic_constraints.isCA;
	return SECSuccess;
}

/* Check if the raw GUID shows up in the string. */
static PRBool
guid_matches(const char *guid, SECItem *raw_guid)
{
	unsigned char bytes[16];
	int i;
	long l;
	const char *p;
	char *q;
	/* We only check properly-formatted GUIDs. */
	if (raw_guid->len != 16) {
		pkinit_debug("GUID claims to have length %d, bogus!\n",
			     raw_guid->len);
		return PR_FALSE;
	}
	/* Now walk the string and extract enough bytes to make a full GUID. */
	p = guid;
	i = 0;
	while (*p != '\0') {
		l = strtol(p, &q, 16);
		if ((q == NULL) || ((q - p) > 2)) {
			/* Malformed. */
			break;
		}
		pkinit_debug(" trusted_guid[%d] = 0x%02x\n", i, l & 0xff);
		bytes[i++] = (l & 0xff);
		p = q + strcspn(q, "0123456789ABCDEFabcdef");
		if (i == 16) {
			if (memcmp(bytes, raw_guid->data, 16) == 0) {
				pkinit_debug("The trusted_guid value "
					     "matches.\n");
				return PR_TRUE;
			}
			i = 0;
			if (*p != '\0') {
				pkinit_debug("More data, will keep trying.\n");
			}
		}
	}
	if (i != 0) {
		pkinit_debug("Discarding partially-listed GUID.\n");
	}

	/* Done! */
	return PR_FALSE;
}

/* Pull out the GUID from the specified certificate. */
SECItem **
cert_get_guids(CERTCertificate *cert)
{
	struct subject_alt_name **names;
	SECItem san_value, guid;
	SECItem **values;
	int i, j;

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug("No subjectAltName extension in certificate!\n");
		return NULL;
	}

	/* Split up the subjectAltName sequence. */
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug("Error parsing subjectAltName extension!\n");
		return NULL;
	}

	/* Iterate over all of the values. */
	for (i = 0; (names != NULL) && (names[i] != NULL); i++) {
		continue;
	}
	values = PORT_ArenaZAlloc(cert->arena, sizeof(SECItem *) * (i + 1));
	if (values == NULL) {
		return NULL;
	}
	for (i = 0, j = 0; (names != NULL) && (names[i] != NULL); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_other_name:
			/* Check for a GUID. */
			if (SECITEM_ItemsAreEqual(&oid_ms_guid.oid,
						  &names[i]->subject_alt_name_value.other_name.oid)) {
				/* Try to decode and store the value. */
				if (SEC_ASN1DecodeItem(cert->arena,
						       &guid,
						       ms_guid_template,
						       &names[i]->subject_alt_name_value.other_name.data) == SECSuccess) {
					values[j++] = SECITEM_ArenaDupItem(cert->arena, &guid);
				}
			}
			break;
		}
	}
	if (j > 0) {
		return values;
	} else {
		return NULL;
	}
}

/* Check if the specified certificate contains a GUID which is listed as one of
 * those which are trusted to hand out TGTs for the server's realm. */
static SECStatus
cert_san_has_guid_for_realm(CERTCertificate *kdc,
			    krb5_context kcontext,
			    krb5_principal server,
			    PRBool *is_trusted)
{
	struct subject_alt_name **names;
	SECItem **guids;
	char *trusted_guid;
	int i;

	i = 0;
	names = NULL;

	/* Find out if we have a list of trusted GUIDs. */
	krb5_appdefault_string(kcontext, "pkinit",
			       krb5_princ_realm(kcontext, server),
			       "trusted_guid", "", &trusted_guid);
	if ((trusted_guid == NULL) || (strcmp(trusted_guid, "") == 0)) {
		pkinit_debug("No configured trusted_guid!\n");
		return SECFailure;
	}

	/* Retrieve the GUID values. */
	guids = cert_get_guids(kdc);
	for (i = 0; (guids != NULL) && (guids[i] != NULL); i++) {
		/* Compare it to the listed values. */
		if (guid_matches(trusted_guid, guids[i])) {
			pkinit_debug("Found matching GUID!\n");
			*is_trusted = PR_TRUE;
		}
	}
	return SECSuccess;
}

/* Determine if the certificate is a valid domain controller certificate for
 * the realm. */
PRBool
cert_is_domain_controller(CERTCertificate *signer,
			  krb5_context kcontext,
			  krb5_principal tgt_principal)
{
	PRBool matches;
	if (((cert_template_name_matches(signer,
					 "DomainController",
					 &matches) != SECSuccess) ||
	     !matches) &&
	    ((cert_template_name_matches(signer,
	   				 "DomainControllerAuthentication",
					 &matches) != SECSuccess) ||
	     !matches)) {
		pkinit_debug("Certificate doesn't belong to a "
			     "domain controller!\n");
		return PR_FALSE;
	}
	if ((cert_san_has_guid_for_realm(signer,
					 kcontext,
					 tgt_principal,
					 &matches) != SECSuccess) ||
	    !matches) {
		pkinit_debug("Certificate GUID is not trusted for realm!\n");
		return PR_FALSE;
	}
	return PR_TRUE;
}

/* Check if we should trust a given certificate which was used by a KDC. */
krb5_error_code
cert_validate_kdc_certificate(CERTCertDBHandle *certdb, CERTCertificate *kdc,
			      struct pkinit_pwcb_args *pwcb_args,
			      krb5_context kcontext, krb5_principal server)
{
	SECCertificateUsage usages;
	PRBool is_ca, is_dc, is_trusted, matches;

	/* Do the default validity check. */
	if (CERT_VerifyCertificateNow(certdb, kdc, PR_TRUE,
				      0,
				      pwcb_args,
				      &usages) != SECSuccess) {
		pkinit_debug("KDC certificate (\"%s\" issued by \"%s\") failed simple validation check!\n", kdc->subjectName, kdc->issuerName);
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Check that it's a terminal node. */
	is_ca = PR_TRUE;
	if ((cert_certificate_get_is_ca(kdc, &is_ca) == SECSuccess) && is_ca) {
		pkinit_debug("KDC certificate shouldn't also be a CA "
			     "certificate!\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* If it has the realm's TGS name as its principal name, and the
	 * certificate can be used for server authentication, then that's good
	 * enough for us. */
	if ((cert_san_matches_principal(kdc, server, &matches) == SECSuccess) &&
	    matches) {
		pkinit_debug("KDC certificate has matching PKINIT SAN.\n");
		/* PKINIT name matched. Check if we need to check EKU stuffs. */
		if (cert_eku_matches_oid(kdc,
					 &oid_pkinit_key_purpose_kdc.oid,
					 &matches) == SECSuccess) {
			if (!matches) {
				/* Oops, it has EKU, but not the one we want. */
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
			if ((cert_ku_matches_mask(kdc, KU_DIGITAL_SIGNATURE,
						  &matches) != SECSuccess) ||
			    !matches) {
				/* Oops, it needs to be able to sign stuff. */
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
		}
		return 0;
	}
	cert_ku_matches_mask(kdc, KU_DIGITAL_SIGNATURE, &matches); /* XXX */

	/* We've got a Windows-style certificate name? */
	if ((cert_san_matches_upn(kdc, kcontext, server,
				  &matches) == SECSuccess) &&
	    matches) {
		if (cert_eku_matches_tag(kdc,
					 SEC_OID_EXT_KEY_USAGE_SERVER_AUTH,
					 &matches) == SECSuccess) {
			if (!matches) {
				/* Oops, it has EKU, but not the one we want. */
				pkinit_debug("KDC certificate can't be used "
					     "for server authentication.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
			if ((cert_ku_matches_mask(kdc, KU_DIGITAL_SIGNATURE,
						  &matches) != SECSuccess) ||
			    !matches) {
				/* Oops, it needs to be able to sign stuff. */
				pkinit_debug("KDC certificate can't be used "
					     "for signing.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
		}
		pkinit_debug("KDC certificate has matching UPN SAN.\n");
		return 0;
	}

	/* Otherwise do the best we can to see if it's a valid Windows
	 * domain controller certificate.  First, check if the
	 * certificateTemplateName has a good value. */
	is_dc = PR_FALSE;
	if (((cert_template_name_matches(kdc, "DomainController",
					 &is_dc) != SECSuccess) ||
	     !is_dc) &&
	    ((cert_template_name_matches(kdc, "DomainControllerAuthentication",
					 &is_dc) != SECSuccess) ||
	     !is_dc)) {
		pkinit_debug("KDC certificate doesn't belong to a "
			     "domain controller!\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Check that it has a GUID, and that the GUID is in the list
	 * which we trust to sign TGTs for the given realm. */
	is_trusted = PR_FALSE;
	if (cert_san_has_guid_for_realm(kdc, kcontext, server,
					&is_trusted) != SECSuccess) {
		pkinit_debug("KDC certificate has no GUID extension!\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}
	if (!is_trusted) {
		pkinit_debug("KDC certificate GUID is not trusted by "
			     "client configuration!\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Done. */
	pkinit_debug("KDC certificate successfully validated!\n");
	return 0;
}

krb5_error_code
cert_validate_client_certificate(CERTCertDBHandle *certdb,
				 CERTCertificate *cert,
			         struct pkinit_pwcb_args *pwcb_args,
				 krb5_context kcontext, krb5_principal client)
{
	SECCertificateUsage usages;
	PRBool is_ca, matches;

	/* Do the default validity check. */
	if (CERT_VerifyCertificateNow(certdb, cert, PR_TRUE,
				      certificateUsageSSLClient,
				      pwcb_args,
				      &usages) != SECSuccess) {
		pkinit_debug("Client certificate failed simple validation check!\n");
		return KRB5KDC_ERR_CLIENT_NOT_TRUSTED;
	}

	/* Check that it's a terminal node. */
	if ((cert_certificate_get_is_ca(cert, &is_ca) == SECSuccess) &&
	    is_ca) {
		pkinit_debug("Client certificate shouldn't be a CA certificate!\n");
		return KRB5KDC_ERR_CLIENT_NOT_TRUSTED;
	}

	/* Check that it matches the principal name. */
	if (!cert_is_preferred(cert, kcontext, client, CERT_MATCHES_UPN_SAN) &&
	    !cert_is_preferred(cert, kcontext, client, CERT_MATCHES_KRB_SAN)) {
		pkinit_debug("Client certificate didn't match client name!\n");
		return KRB5KDC_ERR_CERTIFICATE_MISMATCH;
	}

	/* Check that it matches the desired EKU. */
	if ((cert_eku_matches_oid(cert,
				  &oid_pkinit_key_purpose_client.oid,
				  &matches) == SECSuccess) &&
	    !matches &&
	    (cert_eku_matches_oid(cert,
				  &oid_ms_sc_login.oid,
				  &matches) == SECSuccess) &&
	    !matches) {
		/* Oops, it has EKU, but not the one we want. */
		pkinit_debug("Client certificate can't be used "
			     "for PKINIT authentication.\n");
		return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
	}
	if ((cert_ku_matches_mask(cert, KU_DIGITAL_SIGNATURE,
				  &matches) == SECSuccess) &&
	    !matches) {
		/* Oops, it needs to be able to sign stuff. */
		pkinit_debug("Client certificate can't be used for signing.\n");
		return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
	}

	pkinit_debug("Client certificate has matching SAN.\n");
	return 0;
}
