/*
 * Copyright (C) 2006 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: pkinit.c,v 1.30 2006/11/01 19:20:28 nalin Exp $"

#include "../config.h"

#include <errno.h>
#include <stdio.h>

#include <krb5/krb5.h>
#include <krb5/preauth_plugin.h>

#include <nss.h>
#include <cert.h>
#include <certdb.h>
#include <ocsp.h>
#include <prerror.h>

#include "commont.h"
#include "bcmst.h"
#include "pkinitt.h"
#include "certs.h"
#include "pkinit.h"

#ifdef USE_PAL_BACKPORT
#include "backport-errors.h"
#define preauthentication_client_0 preauthentication_client_0_backport18750
#define preauthentication_server_0 preauthentication_server_0_backport18750
#endif

#ifndef KRB5_PADATA_PK_AS_REQ_OLD
#undef  KRB5_PADATA_PK_AS_REQ
#define KRB5_PADATA_PK_AS_REQ_OLD	14
#define KRB5_PADATA_PK_AS_REQ		16
#endif
#ifndef KRB5_PADATA_PK_AS_REP_OLD
#undef  KRB5_PADATA_PK_AS_REP
#define KRB5_PADATA_PK_AS_REP_OLD	15
#define KRB5_PADATA_PK_AS_REP		17
#endif

#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#define _(_x) dgettext(PACKAGE_NAME, _x)
#else
#define _(_x) _x
#endif

#if DEBUG
#include <stdarg.h>
void
pkinit_debug(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}
#endif

static krb5_preauthtype supported_client_pa_types[] = {
	KRB5_PADATA_PK_AS_REP_OLD,
	KRB5_PADATA_PK_AS_REQ, KRB5_PADATA_PK_AS_REP,
	0,
};
static krb5_preauthtype supported_server_pa_types[] = {
	KRB5_PADATA_PK_AS_REQ,
	0,
};

struct client_context {
	PLArenaPool *pool;
	SECKEYPrivateKey *dh_private_key;
	SECKEYPublicKey *dh_public_key;
	SECItem *dh_nonce;
};

static void
initialize_translations(void)
{
#ifdef HAVE_LIBINTL_H
	bindtextdomain(PACKAGE_NAME, LOCALEDATADIR);
#endif
}

char *
pkinit_pwcb(PK11SlotInfo *slot, PRBool retry, void *arg)
{
	int ret;
	struct pkinit_pwcb_args *args;
	krb5_prompt prompt;
	krb5_data reply;
	char *name, *text, *answer;
	void *data;

	/* We only want to be called once. */
	if (retry) {
		pkinit_debug("Prompter callback called more than once, "
			     "failing\n");
		return NULL;
	}
	/* We need our callback arguments. */
	if (arg == NULL) {
		pkinit_debug("Prompter callback called with no args!\n");
		return NULL;
	}
	args = arg;
	if (args->kcontext == NULL) {
		pkinit_debug("Prompter callback called with no Kerberos "
			     "context!\n");
		return NULL;
	}
	if (args->fct == NULL) {
		pkinit_debug("Prompter callback called with no prompter!\n");
		return NULL;
	}
	if (args->pool == NULL) {
		pkinit_debug("Prompter callback called with no pool!\n");
		return NULL;
	}

	/* Get the name of the token. */
	name = PK11_GetTokenName(slot);
	if (name == NULL) {
		pkinit_debug("Skipping nameless token.\n");
		return NULL;
	}

	/* Set up the prompt. */
	text = PORT_ArenaZAlloc(args->pool, strlen(name) + 100);
	snprintf(text, strlen(name) + 100, _("PIN for %s"), name);
	memset(&prompt, 0, sizeof(prompt));
	prompt.prompt = text;
	prompt.hidden = 1;
	prompt.reply = &reply;
	reply.length = 256;
	data = malloc(reply.length);
	reply.data = data;
	name = NULL;
	answer = NULL;

#if DEBUG
	name = "PKINIT-NSS";
	answer = PACKAGE_STRING;
#endif
	/* Call the callback. */
	ret = (*args->fct)(args->kcontext, args->fct_data,
			   name, answer, 1, &prompt);
	answer = NULL;

	if ((ret == 0) && (reply.data != NULL)) {
		answer = PR_Malloc(reply.length + 1);
		memcpy(answer, reply.data, reply.length);
		answer[reply.length] = '\0';
	}

	if (reply.data == data) {
		free(reply.data);
	}

	return answer;
}

/* Initialize NSS, enable/disable OCSP checking, start up i18n support. */
static krb5_error_code
initialize_nss(krb5_context kcontext,
	       krb5_principal principal,
	       const char *dbdir,
	       PRBool as_server,
	       CERTCertDBHandle **certdb)
{
	int allowed;

	/* Check if we're disabled. */
	pkinit_debug("Checking if PKINIT is allowed.\n");
	allowed = TRUE;
	krb5_appdefault_boolean(kcontext, "pkinit",
				krb5_princ_realm(kcontext, principal),
				"allow_pkinit", TRUE, &allowed);
	if (allowed) {
		pkinit_debug("PKINIT is allowed.\n");
	} else {
		pkinit_debug("PKINIT is disallowed.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Initialize the library. */
	if (!NSS_IsInitialized()) {
		SECStatus ret = NSS_Init(dbdir);
		if (ret != SECSuccess) {
			char *err = NULL;
			PRInt32 errlen = PR_GetErrorTextLength();

			if (errlen) {
				err = malloc(errlen + 1);
			}
			errlen = PR_GetErrorText(err);
			pkinit_debug("NSS_Init returned error %d(%s)\n",
				     ret, errlen ? err : "none");
			if (err) {
				free(err);
			}
			return EINVAL;
		}
	}
	*certdb = CERT_GetDefaultCertDB();
	if (*certdb == NULL) {
		pkinit_debug("No NSS database found.\n");
		NSS_Shutdown();
		return EINVAL;
	}

	/* Turn OCSP checking on or off. */
	krb5_appdefault_boolean(kcontext, "pkinit",
				krb5_princ_realm(kcontext, principal),
				"ocsp_checking",
				as_server ? PR_TRUE : PR_FALSE,
				&allowed);
	if (allowed) {
		CERT_EnableOCSPChecking(*certdb);
	} else {
		CERT_DisableOCSPChecking(*certdb);
	}
	initialize_translations();

	return 0;
}

static krb5_error_code
client_init(krb5_context kcontext, krb5_preauthtype pa_type,
	    void **module_context)
{
	PLArenaPool *pool;
	struct client_context *ctx;
	if (*module_context == NULL) {
		pkinit_debug("Creating new context.\n");
		pool = PORT_NewArena(sizeof(double));
		if (pool == NULL) {
			return ENOMEM;
		}
		ctx = PORT_ArenaZAlloc(pool, sizeof(*ctx));
		ctx->pool = pool;
		*module_context = ctx;
	}
	return 0;
}

static void
client_fini(krb5_context kcontext, krb5_preauthtype pa_type,
	    void *module_context)
{
	PLArenaPool *pool;
	struct client_context *ctx;

	if (module_context) {
		/* Save a pointer to the pool. */
		ctx = module_context;
		pool = ctx->pool;
		/* Zero out the structure. */
		memset(ctx, 0, sizeof(*ctx));
		/* Free the pool. */
		PORT_FreeArena(pool, PR_TRUE);
	}
}

static krb5_error_code
client_process(krb5_context kcontext,
	       void *module_context,
	       void **client_context,
	       krb5_kdc_req *request,
	       krb5_data *encoded_request_body,
	       krb5_data *encoded_previous_request,
	       krb5_pa_data *pa_data,
	       krb5_prompter_fct prompter,
	       void *prompter_data,
	       preauth_get_as_key_proc gak_fct,
	       krb5_data *salt, krb5_data *s2kparams,
	       void *gak_data,
	       krb5_keyblock *as_key,
	       krb5_pa_data **out_pa_data)
{
	krb5_data *req, rep;
	krb5_pa_data *pa_req;
	krb5_error_code ret;
	CERTCertDBHandle *certdb;
	CERTCertificate *cert = NULL;
	PK11SlotInfo *preferred_slot;
	struct pkinit_pwcb_args pwcb_args;
	const char *preferred_slot_name;
	struct client_context *ctx;

	*out_pa_data = NULL;
	preferred_slot = NULL;
	ret = KRB5KRB_ERR_GENERIC;

	/* Hey! No cheating! Either we have a challenge or a response. */
	if (pa_data == NULL) {
		return EINVAL;
	}

	/* Do the shared initialization stuff. */
	ret = initialize_nss(kcontext, request->client,
			     DEFAULT_PKINIT_CLIENT_DBDIR, PR_FALSE, &certdb);
	if (ret != 0) {
		return ret;
	}

	/* Recover the module-wide context. */
	ctx = module_context;

	/* Set up callbacks for the library in case it needs to ask the user
	 * for a PIN. */
	pwcb_args.kcontext = kcontext;
	pwcb_args.fct = prompter;
	pwcb_args.fct_data = prompter_data;
	pwcb_args.pool = ctx->pool;
	PK11_SetPasswordFunc(pkinit_pwcb);

	/* If we have a preferred slot, try to get a handle for it. */
	preferred_slot_name = getenv("PKCS11_LOGIN_TOKEN_NAME");
	if (preferred_slot_name != NULL) {
		preferred_slot = PK11_FindSlotByName(preferred_slot_name);
	}

	/* Try to find the preferred certificate for the user. */
	pkinit_debug("Looking for the user's certificate.\n");
	cert = cert_find_preferred_cert(preferred_slot, &pwcb_args,
					kcontext, request->client,
					CERT_MATCHES_KRB_SAN |
					CERT_MATCHES_UPN_SAN);
	if (cert == NULL) {
		pkinit_debug("Unable to find certificate.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}
	pkinit_debug("Using \"%s\".\n", cert->nickname);

	switch (pa_data->pa_type) {
	case KRB5_PADATA_PK_AS_REP_OLD:
		if (pa_data->length == 0) {
			/* Create a new draft pa_pk_as_req structure, sign it
			 * with a matching certificate, and prepare it for
			 * transmission. */
			pkinit_debug("Called to create a draft request.\n");
			req = pkinit_create_draft_pa_pk_as_req(ctx->pool,
							       kcontext,
							       request,
							       certdb,
							       cert,
							       preferred_slot,
							       &ctx->dh_private_key,
							       &ctx->dh_public_key,
							       &ctx->dh_nonce,
							       &pwcb_args);
			/* If that succeeded, wrap the result up in a pa_data
			 * structure. */
			if (req != NULL) {
				pa_req = malloc(sizeof(krb5_pa_data));
				/* Aaagh! WHY is this not _REQ? */
				pa_req->pa_type = KRB5_PADATA_PK_AS_REP_OLD;
				pa_req->length = req->length;
				pa_req->contents = malloc(req->length);
				memcpy(pa_req->contents, req->data,
				       req->length);
				*out_pa_data = pa_req;
				/* Note success for our caller. */
				ret = 0;
			}
		} else {
			pkinit_debug("Called to handle draft response.\n");
			/* The server supplied data, so assume it's a response
			 * and process it. */
			rep.data = (char *) pa_data->contents;
			rep.length = pa_data->length;
			ret = pkinit_verify_draft_pa_pk_as_rep(ctx->pool,
							       certdb,
							       preferred_slot,
							       &pwcb_args,
							       kcontext,
							       request,
							       &rep,
							       as_key);
		}
		break;
	case KRB5_PADATA_PK_AS_REQ:
		/* Create a new pa_pk_as_req structure, sign it
		 * with a matching certificate, and prepare it for
		 * transmission. */
		pkinit_debug("Called to create a RFC4556 request.\n");
		req = pkinit_create_pa_pk_as_req(ctx->pool,
						 kcontext,
						 request,
						 encoded_request_body,
						 certdb,
						 cert,
						 SEC_OID_SHA1,
						 preferred_slot,
						 &ctx->dh_private_key,
						 &ctx->dh_public_key,
						 &ctx->dh_nonce,
						 &pwcb_args);
		/* If that succeeded, wrap the result up in a pa_data
		 * structure. */
		if (req != NULL) {
			pa_req = malloc(sizeof(krb5_pa_data));
			pa_req->pa_type = KRB5_PADATA_PK_AS_REQ;
			pa_req->length = req->length;
			pa_req->contents = malloc(req->length);
			memcpy(pa_req->contents, req->data, req->length);
			*out_pa_data = pa_req;
			/* Note success for our caller. */
			ret = 0;
		}
		break;
	case KRB5_PADATA_PK_AS_REP:
		pkinit_debug("Called to handle RFC4556 response.\n");
		/* The server supplied data -> process the response. */
		rep.data = (char *) pa_data->contents;
		rep.length = pa_data->length;
		/* Verify the result. */
		ret = pkinit_verify_pa_pk_as_rep(ctx->pool,
						 certdb,
						 preferred_slot,
						 &pwcb_args,
						 kcontext,
						 request,
						 encoded_previous_request,
						 &rep,
						 ctx->dh_private_key,
						 ctx->dh_public_key,
						 ctx->dh_nonce,
						 as_key);
		break;
	default:
		break;
	}

	if (ret == 0) {
		pkinit_debug("Returning success to libkrb5.\n");
	} else {
		pkinit_debug("Returning failure (%s) to libkrb5.\n",
			     error_message(ret));
	}
	return ret;
}

static int
client_get_flags(krb5_context kcontext, krb5_preauthtype patype)
{
	return PA_REAL;
}

/* Server-side context. */
struct pa_context {
	krb5_data *client_request;
	SECItem *client_cert;
	SECItem *client_public_info;
	SECItem *client_nonce;
	SECItem *client_dh_nonce;
};

/* Free state. */
static krb5_error_code
server_free_pa_context(krb5_context kcontext,
		       void *module_context, void **pa_context)
{
	struct pa_context *context;

	if (pa_context != NULL) {
		context = *pa_context;
		if (context != NULL) {
			if (context->client_cert != NULL) {
				SECITEM_FreeItem(context->client_cert, PR_TRUE);
				context->client_cert = NULL;
			}
			if (context->client_request != NULL) {
				krb5_free_data(kcontext,
					       context->client_request);
				context->client_request = NULL;
			}
			free(context);
		}
		*pa_context = NULL;
	}

	return 0;
}

/* Obtain and return any preauthentication data (which is destined for the
 * client) which matches type data->pa_type. */
static krb5_error_code
server_get_edata(krb5_context kcontext,
		 krb5_kdc_req *request,
		 struct _krb5_db_entry_new *client,
		 struct _krb5_db_entry_new *server,
	         preauth_get_entry_data_proc server_get_entry_data,
		 void *module_context,
		 krb5_pa_data *data)
{
	CERTCertificate *cert;
	CERTCertDBHandle *certdb;
	PLArenaPool *pool;
	krb5_principal tgs_principal;
	krb5_error_code ret;

	/* No data, but double-check that we do in fact support this type of
	 * preauthentication. */
	if (data->pa_type != KRB5_PADATA_PK_AS_REQ) {
		/* Any non-zero value is treated as an error, so this works. */
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Initialize the library. */
	ret = initialize_nss(kcontext, request->server,
			     DEFAULT_PKINIT_SERVER_DBDIR, PR_TRUE, &certdb);
	if (ret != 0) {
		return ret;
	}

	/* If we can't create a new memory arena, bail. */
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		pkinit_debug("Out of memory.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Build the KDC's principal name. */
	tgs_principal = NULL;
	if (pkinit_build_tgs_name(kcontext, request->server,
				  &tgs_principal) != 0) {
		pkinit_debug("Error building TGS principal name.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Check if we have a KDC certificate for the server, and error out if
	 * one isn't found, to prevent the KDC from listing us as a supported
	 * preauth type when we can't really do anything. */
	pkinit_debug("Looking for the KDC's certificate.\n");
	cert = cert_find_preferred_cert(NULL, NULL,
					kcontext, tgs_principal,
					CERT_MATCHES_KRB_SAN |
					CERT_MATCHES_UPN_SAN);

	/* Clean up. */
	krb5_free_principal(kcontext, tgs_principal);
	PORT_FreeArena(pool, PR_TRUE);

	/* We're not really concerned with the contents of the certificate at
	 * this point, merely that we found one which we can use. */
	if (cert == NULL) {
		pkinit_debug("No suitable certificate found.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}
	pkinit_debug("Found a suitable certificate.\n");

	return 0;
}

/* Add a member to a list of pointers. */
static void **
make_pointer_list(void **current_list, void *new_member)
{
	int i;
	void **ret;
	if (current_list == NULL) {
		ret = malloc(2 * sizeof(void*));
		if (ret != NULL) {
			ret[0] = new_member;
			ret[1] = NULL;
		}
		return ret;
	}
	for (i = 0; current_list[i] != NULL; i++) {
		continue;
	}
	ret = malloc((i + 2) * sizeof(void*));
	if (ret != NULL) {
		for (i = 0; current_list[i] != NULL; i++) {
			ret[i] = current_list[i];
		}
		ret[i++] = new_member;
		ret[i++] = NULL;
		free(current_list);
	}
	return ret;
}

/* Verify a request from a client. */
static krb5_error_code
server_verify(krb5_context kcontext,
	      struct _krb5_db_entry_new *client,
	      krb5_data *req_pkt,
	      krb5_kdc_req *request,
	      krb5_enc_tkt_part *enc_tkt_reply,
	      krb5_pa_data *data,
	      preauth_get_entry_data_proc server_get_entry_data,
	      void *pa_module_context,
	      void **pa_request_context,
	      krb5_data **e_data)
{
	PLArenaPool *pool;
	krb5_error_code ret;
	struct pa_context *context;
	krb5_data *client_entry_cert, *client_max_skew, preauth_data;
	krb5_authdata *authdata;
	krb5_deltat max_skew;
	SECItem client_cert_item, *mapped_certs, *item;
	int n_mapped_certs;
	CERTCertDBHandle *certdb;
	CERTCertificate *signer;
	struct subject_public_key_info *client_public_info;
	SECItem *client_nonce, *client_dh_nonce;

	/* Initialize the library. */
	ret = initialize_nss(kcontext, request->server,
			     DEFAULT_PKINIT_SERVER_DBDIR, PR_TRUE, &certdb);
	if (ret != 0) {
		return ret;
	}

	/* Allocate the server context, but do it using the heap. */
	context = malloc(sizeof(struct pa_context));
	if (context == NULL) {
		return ENOMEM;
	}
	*pa_request_context = context;
	memset(context, 0, sizeof(*context));

	/* If we can't create a new memory arena, bail. */
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		return ENOMEM;
	}

	/* Determine the maximum clock skew. */
	client_max_skew = NULL;
	ret = (*server_get_entry_data)(kcontext, request, client,
				       krb5plugin_preauth_entry_max_time_skew,
				       &client_max_skew);
	if ((ret != 0) || (client_max_skew->length != sizeof(max_skew))) {
		pkinit_debug("Unable to determine max permitted clock skew.\n");
		return ret;
	}
	max_skew = *(krb5_deltat *)(client_max_skew->data);

	/* Check if the client has an authorized certificate in the database. */
	client_entry_cert = NULL;
	memset(&client_cert_item, 0, sizeof(client_cert_item));
	mapped_certs = NULL;
	ret = (*server_get_entry_data)(kcontext, request, client,
				       krb5plugin_preauth_entry_max_time_skew,
				       &client_entry_cert);
	if (client_entry_cert != NULL) {
		client_cert_item.data =
			(unsigned char *) client_entry_cert->data;
		client_cert_item.len = client_entry_cert->length;
		mapped_certs = &client_cert_item;
		n_mapped_certs = 1;
	}

	/* Verify the request. */
	preauth_data.data = (char *) data->contents;
	preauth_data.length = data->length;
	signer = NULL;
	client_public_info = NULL;
	client_nonce = NULL;
	client_dh_nonce = NULL;
	authdata = NULL;
	ret = pkinit_verify_pa_pk_as_req(pool, certdb,
					 kcontext, request, req_pkt,
					 &preauth_data, max_skew,
					 &mapped_certs,
					 n_mapped_certs,
					 &signer,
					 &client_public_info,
					 &client_nonce,
					 &client_dh_nonce,
					 &authdata);
	if (ret == 0) {
		/* Success!  Note that we count as preauthentication. */
		enc_tkt_reply->flags |= TKT_FLG_PRE_AUTH;

		/* We're hardware preauthentication, too.
		 * FIXME: _we_'re using smart cards, but provide a way to turn
		 * this off for admins who know better about their deployments.
		 */
		enc_tkt_reply->flags |= TKT_FLG_HW_AUTH;

		/* Cache a copy of the client's certificate. */
		context->client_cert = &signer->derCert;

		/* Add supplied authdata to the list. */
		if (authdata != NULL) {
			enc_tkt_reply->authorization_data = (krb5_authdata**)make_pointer_list((void**) enc_tkt_reply->authorization_data, authdata);
		}

		/* The certificate is "owned" by the pool, but we need
		 * a copy which will survive the pool's destruction. */
		if (context->client_cert != NULL) {
			context->client_cert =
				SECITEM_DupItem(context->client_cert);
			context->client_request =
				malloc(sizeof(*context->client_request));
			krb5_copy_data(kcontext, req_pkt,
				       &context->client_request);
		}

		/* Save the client's DH info. */
		if (client_public_info != NULL) {
			item = common_encode_subject_public_key_info(pool,
								     client_public_info);
			context->client_public_info = SECITEM_DupItem(item);
			pkinit_debug("Saved client DH key information.\n");
		}
		if (client_dh_nonce != NULL) {
			context->client_dh_nonce =
				SECITEM_DupItem(client_dh_nonce);
			pkinit_debug("Saved client DH nonce.\n");
		}

		/* Save the client's nonce. */
		if (client_nonce != NULL) {
			context->client_nonce = SECITEM_DupItem(client_nonce);
			pkinit_debug("Saved client nonce.\n");
		}
	}

	/* Clean up. */
	PORT_FreeArena(pool, PR_TRUE);

	return ret;
}

/* Create the response for a client. */
static krb5_error_code
server_return(krb5_context kcontext,
	      krb5_pa_data *padata,
	      struct _krb5_db_entry_new *client,
	      krb5_data *req_pkt,
	      krb5_kdc_req *request,
	      krb5_kdc_rep *reply,
	      struct _krb5_key_data *client_key,
	      krb5_keyblock *encrypting_key,
	      krb5_pa_data **send_pa,
	      preauth_get_entry_data_proc server_get_entry_data,
	      void *pa_module_context,
	      void **pa_request_context)
{
	PLArenaPool *pool;
	CERTCertDBHandle *certdb;
	struct pa_context *context;
	krb5_data *rep;
	krb5_pa_data *pa_data;
	krb5_error_code ret;
	struct subject_public_key_info *client_public_info;

	ret = ENOMEM;

	/* We can't return data if we didn't process a PKINIT preauth request.
	 */
	if ((pa_request_context == NULL) || (*pa_request_context == NULL)) {
		return 0;
	}
	context = *pa_request_context;

	/* Initialize the library. */
	ret = initialize_nss(kcontext, request->server,
			     DEFAULT_PKINIT_SERVER_DBDIR, PR_TRUE, &certdb);
	if (ret != 0) {
		return ret;
	}

	/* If we can't create a new memory arena, bail. */
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		return ENOMEM;
	}

	/* Build a new reply structure. */
	if ((context->client_cert != NULL) &&
	    (context->client_request != NULL)) {
		client_public_info = NULL;
		if (context->client_public_info != NULL) {
			client_public_info = common_decode_subject_public_key_info(pool,
										   context->client_public_info);
		}
		rep = pkinit_create_pa_pk_as_rep(pool,
						 certdb,
						 kcontext,
						 request,
						 encrypting_key,
						 context->client_request,
						 context->client_cert,
						 client_public_info,
						 context->client_dh_nonce);
		if (rep != NULL) {
			pa_data = malloc(sizeof(*pa_data));
			if (pa_data == NULL) {
				pkinit_debug("Out of memory.\n");
			} else {
				pa_data->pa_type = KRB5_PADATA_PK_AS_REP;
				pa_data->length = rep->length;
				pa_data->contents = malloc(rep->length);
				if (pa_data->contents == NULL) {
					pkinit_debug("Out of memory.\n");
					free(pa_data);
				} else {
					memcpy(pa_data->contents,
					       rep->data, rep->length);
					*send_pa = pa_data;
					pkinit_debug("Generated response.\n");
					ret = 0;
				}
			}
		}
	}

	/* Clean up. */
	PORT_FreeArena(pool, PR_TRUE);

	/* Clear the context. */
	server_free_pa_context(kcontext, pa_module_context, pa_request_context);

	pkinit_debug("Returning %d (%s).\n", ret, error_message(ret));
	return ret;
}

static int
server_get_flags(krb5_context kcontext, krb5_preauthtype patype)
{
	return PA_SUFFICIENT | PA_REPLACES_KEY;
}

struct krb5plugin_preauth_client_ftable_v0 preauthentication_client_0 = {
	.name = "pkinit-nss",
	.pa_type_list = &supported_client_pa_types[0],
	.enctype_list = NULL,
	.init = client_init,
	.fini = client_fini,
	.process = client_process,
	.flags = client_get_flags,
};
struct krb5plugin_preauth_server_ftable_v0 preauthentication_server_0 = {
	.name = "pkinit-nss",
	.pa_type_list = &supported_server_pa_types[0],
	.flags_proc = server_get_flags,
	.edata_proc = server_get_edata,
	.verify_proc = server_verify,
	.return_proc = server_return,
	.freepa_reqcontext_proc = server_free_pa_context,
};
