/*
 * Copyright (C) 2006 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: pkinitt.c,v 1.19 2006/11/01 16:32:59 nalin Exp $"

#include "../config.h"

#include <krb5.h>
#include <nss.h>
#include <pk11pub.h>
#include <pk11pqg.h>
#include <time.h>

#include "commont.h"
#include "pkinitt.h"
#include "certs.h"
#include "oakley.h"

#ifdef USE_PAL_BACKPORT
#include "backport-errors.h"
#endif

static void
derdump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}
static void
dump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("od -t x1c", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}

static unsigned char pkinit_null_bytes[] = {SEC_ASN1_NULL, 0};
static SECItem pkinit_null_item = {
	.data = &pkinit_null_bytes[0],
	.len = 2,
};

static const SEC_ASN1Template
pkinit_realm_template[] = {
	{
	.kind = SEC_ASN1_GENERAL_STRING,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(SECItem),
	},
};

static const SEC_ASN1Template
pkinit_kerberos_name_string_template[] = {
	{
	.kind = SEC_ASN1_GENERAL_STRING,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(SECItem),
	},
};

static const SEC_ASN1Template
pkinit_sequence_of_kerberos_name_string_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = &pkinit_kerberos_name_string_template,
	.size = 0,
	},
};

static const SEC_ASN1Template
pkinit_principal_name_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct principal_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct principal_name, name_type),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct principal_name, name_string),
	.sub = pkinit_sequence_of_kerberos_name_string_template,
	.size = sizeof(struct SECItem**),
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
pkinit_kerberos_name_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct kerberos_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kerberos_name, realm),
	.sub = &pkinit_realm_template,
	.size = sizeof(struct realm),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kerberos_name, principal_name),
	.sub = &pkinit_principal_name_template,
	.size = sizeof(struct principal_name),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_draft_pkauthenticator_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct draft_pkauthenticator),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pkauthenticator, kdc_name),
	.sub = pkinit_principal_name_template,
	.size = sizeof(struct principal_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pkauthenticator, kdc_realm),
	.sub = pkinit_realm_template,
	.size = sizeof(struct realm),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pkauthenticator, cusec),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 3 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pkauthenticator, ctime),
	.sub = &SEC_GeneralizedTimeTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 4 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pkauthenticator, nonce),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_draft_auth_pack_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct draft_auth_pack),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_auth_pack, pkauthenticator),
	.sub = pkinit_draft_pkauthenticator_template,
	.size = sizeof(struct draft_pkauthenticator),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct draft_auth_pack, client_public_value),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_pkauthenticator_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct pkauthenticator),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct pkauthenticator, cusec),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct pkauthenticator, ctime),
	.sub = &SEC_GeneralizedTimeTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct pkauthenticator, nonce),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 3 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct pkauthenticator, pa_checksum),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_auth_pack_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct auth_pack),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct auth_pack, pkauthenticator),
	.sub = pkinit_pkauthenticator_template,
	.size = sizeof(struct pkauthenticator),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct auth_pack, client_public_value),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct auth_pack, supported_cms_types),
	.sub = common_sequence_of_algorithm_identifier_template,
	.size = sizeof(struct algorithm_identifier **),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 3 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct auth_pack, client_dh_nonce),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_trusted_cas_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct trusted_cas, trusted_ca_type),
	.sub = NULL,
	.size = sizeof(struct trusted_cas),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct trusted_cas, trusted_ca_value.principal_name),
	.sub = &pkinit_principal_name_template,
	.size = trusted_ca_type_principal_name,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct trusted_cas, trusted_ca_value.name),
	.sub = &SEC_AnyTemplate,
	.size = trusted_ca_type_name,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct trusted_cas,
			   trusted_ca_value.issuer_and_serial),
	.sub = &common_issuer_and_serial_number_template,
	.size = trusted_ca_type_issuer_and_serial_number,
	},
	{0, 0, NULL, 0,},
};

static const SEC_ASN1Template
pkinit_sequence_of_trusted_cas_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = pkinit_trusted_cas_template,
	.size = 0,
	},
};

static const SEC_ASN1Template
pkinit_draft_pa_pk_as_req_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct draft_pa_pk_as_req),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pa_pk_as_req, signed_auth_pack),
	.sub = bcms_constructed_content_info_template,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct draft_pa_pk_as_req, trusted_cas),
	.sub = pkinit_sequence_of_trusted_cas_template,
	.size = sizeof(struct trusted_ca **),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_OPTIONAL |
		SEC_ASN1_POINTER,
	.offset = offsetof(struct draft_pa_pk_as_req, kdc_cert),
	.sub = common_issuer_and_serial_number_template,
	.size = sizeof(struct issuer_and_serial_number),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 3 | SEC_ASN1_OPTIONAL |
		SEC_ASN1_POINTER,
	.offset = offsetof(struct draft_pa_pk_as_req, encryption_cert),
	.sub = common_issuer_and_serial_number_template,
	.size = sizeof(struct issuer_and_serial_number *),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_pa_pk_as_req_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct pa_pk_as_req),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0,
	.offset = offsetof(struct pa_pk_as_req, signed_auth_pack),
	.sub = bcms_constructed_content_info_template,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct pa_pk_as_req, trusted_certifiers),
	.sub = bcms_sequence_of_external_principal_identifiers_template,
	.size = sizeof(struct external_principal_identifiers **),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_OPTIONAL |
		SEC_ASN1_POINTER,
	.offset = offsetof(struct pa_pk_as_req, kdc_pk_id),
	.sub = bcms_signer_identifier_template,
	.size = sizeof(struct signer_identifier*),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_draft_pa_pk_as_rep_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct draft_pa_pk_as_rep, draft_pa_pk_as_rep_type),
	.sub = NULL,
	.size = sizeof(struct draft_pa_pk_as_rep),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pa_pk_as_rep,
			   draft_pa_pk_as_rep_value.enc_key_pack),
	.sub = bcms_constructed_content_info_template,
	.size = draft_pa_pk_as_rep_type_enc_key_pack,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_pa_pk_as_rep,
			   draft_pa_pk_as_rep_value.dh_signed_data),
	.sub = bcms_constructed_content_info_template,
	.size = draft_pa_pk_as_rep_type_dh_signed_data,
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_kdc_dh_key_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct kdc_dh_key_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kdc_dh_key_info, subject_public_key),
	.sub = &SEC_BitStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct kdc_dh_key_info, nonce),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct kdc_dh_key_info, dh_key_expiration),
	.sub = &SEC_GeneralizedTimeTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_dh_rep_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct dh_rep_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0,
	.offset = offsetof(struct dh_rep_info, dh_signed_data),
	.sub = bcms_constructed_content_info_template,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_POINTER | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct dh_rep_info, server_dh_nonce),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_pa_pk_as_rep_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct pa_pk_as_rep, pa_pk_as_rep_type),
	.sub = NULL,
	.size = sizeof(struct pa_pk_as_rep),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1,
	.offset = offsetof(struct pa_pk_as_rep,
			   pa_pk_as_rep_value.enc_key_pack),
	.sub = bcms_constructed_content_info_template,
	.size = pa_pk_as_rep_type_enc_key_pack,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct pa_pk_as_rep,
			   pa_pk_as_rep_value.dh_info),
	.sub = pkinit_dh_rep_info_template,
	.size = pa_pk_as_rep_type_dh_info,
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_encryption_key_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct encryption_key),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct encryption_key, keytype),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct encryption_key, keyvalue),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_draft_reply_key_pack_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct draft_reply_key_pack),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_reply_key_pack, reply_key),
	.sub = pkinit_encryption_key_template,
	.size = sizeof(struct encryption_key),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct draft_reply_key_pack, nonce),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_checksum_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct checksum),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct checksum, cksumtype),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct checksum, checksum),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(struct checksum),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_reply_key_pack_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct reply_key_pack),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct reply_key_pack, reply_key),
	.sub = pkinit_encryption_key_template,
	.size = sizeof(struct encryption_key),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct reply_key_pack, as_checksum),
	.sub = pkinit_checksum_template,
	.size = sizeof(struct checksum),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_kdc_req_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct kdc_req),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kdc_req, pvno),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kdc_req, msg_type),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 3 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kdc_req, padata),
	.sub = common_sequence_of_any_template,
	.size = sizeof(SECItem **),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 4 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT | SEC_ASN1_ANY,
	.offset = offsetof(struct kdc_req, req_body),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};
static const SEC_ASN1Template
pkinit_kdc_as_req_template[] = {
	{
	.kind = SEC_ASN1_APPLICATION | 10 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = 0,
	.sub = pkinit_kdc_req_template,
	.size = sizeof(struct kdc_req),
	},
	{0, 0, NULL, 0},
};

static const SEC_ASN1Template
pkinit_authorization_data_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct authorization_data),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 | SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct authorization_data, ad_type),
	.sub = SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1,
	.offset = offsetof(struct authorization_data, ad_data),
	.sub = SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

struct draft_auth_pack *
pkinit_decode_draft_auth_pack(PLArenaPool *pool, SECItem *item)
{
	struct draft_auth_pack *auth_pack;

	auth_pack = PORT_ArenaZAlloc(pool, sizeof(*auth_pack));
	if (auth_pack == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, auth_pack, pkinit_draft_auth_pack_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding draft auth_pack!\n");
		return NULL;
	}
	return auth_pack;
}

struct auth_pack *
pkinit_decode_auth_pack(PLArenaPool *pool, SECItem *item)
{
	struct auth_pack *auth_pack;

	auth_pack = PORT_ArenaZAlloc(pool, sizeof(*auth_pack));
	if (auth_pack == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, auth_pack, pkinit_auth_pack_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding auth_pack!\n");
		return NULL;
	}
	return auth_pack;
}

struct draft_reply_key_pack *
pkinit_decode_draft_reply_key_pack(PLArenaPool *pool, SECItem *item)
{
	struct draft_reply_key_pack *reply_key_pack;

	reply_key_pack = PORT_ArenaZAlloc(pool, sizeof(*reply_key_pack));
	if (reply_key_pack == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, reply_key_pack,
			       pkinit_draft_reply_key_pack_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding draft reply_key_pack!\n");
		return NULL;
	}
	return reply_key_pack;
}

struct reply_key_pack *
pkinit_decode_reply_key_pack(PLArenaPool *pool, SECItem *item)
{
	struct reply_key_pack *reply_key_pack;

	reply_key_pack = PORT_ArenaZAlloc(pool, sizeof(*reply_key_pack));
	if (reply_key_pack == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, reply_key_pack,
			       pkinit_reply_key_pack_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding reply_key_pack!\n");
		return NULL;
	}
	return reply_key_pack;
}

struct kdc_dh_key_info *
pkinit_decode_kdc_dh_key_info(PLArenaPool *pool, SECItem *item)
{
	struct kdc_dh_key_info *kdc_dh_key_info;

	kdc_dh_key_info = PORT_ArenaZAlloc(pool, sizeof(*kdc_dh_key_info));
	if (kdc_dh_key_info == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, kdc_dh_key_info,
			       pkinit_kdc_dh_key_info_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding kdc_dh_key_info!\n");
		return NULL;
	}
	return kdc_dh_key_info;
}

struct draft_pa_pk_as_req *
pkinit_decode_draft_pa_pk_as_req(PLArenaPool *pool, SECItem *item)
{
	struct draft_pa_pk_as_req *req;

	req = PORT_ArenaZAlloc(pool, sizeof(*req));
	if (req == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, req, pkinit_draft_pa_pk_as_req_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding draft PA-PK-AS-REQ!\n");
		return NULL;
	}
	return req;
}

struct pa_pk_as_req *
pkinit_decode_pa_pk_as_req(PLArenaPool *pool, SECItem *item)
{
	struct pa_pk_as_req *req;

	req = PORT_ArenaZAlloc(pool, sizeof(*req));
	if (req == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, req, pkinit_pa_pk_as_req_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding PA-PK-AS-REQ!\n");
		return NULL;
	}
	return req;
}

struct draft_pa_pk_as_rep *
pkinit_decode_draft_pa_pk_as_rep(PLArenaPool *pool, SECItem *item)
{
	struct draft_pa_pk_as_rep *rep;

	rep = PORT_ArenaZAlloc(pool, sizeof(*rep));
	if (rep == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, rep, pkinit_draft_pa_pk_as_rep_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding draft PA-PK-AS-REQ!\n");
		return NULL;
	}
	return rep;
}

struct pa_pk_as_rep *
pkinit_decode_pa_pk_as_rep(PLArenaPool *pool, SECItem *item)
{
	struct pa_pk_as_rep *rep;

	rep = PORT_ArenaZAlloc(pool, sizeof(*rep));
	if (rep == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, rep, pkinit_pa_pk_as_rep_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding PA-PK-AS-REQ!\n");
		return NULL;
	}
	return rep;
}

struct kerberos_name *
pkinit_decode_kerberos_name(PLArenaPool *pool, SECItem *item)
{
	struct kerberos_name *rep;

	rep = PORT_ArenaZAlloc(pool, sizeof(*rep));
	if (rep == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, rep, pkinit_kerberos_name_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding Kerberos name!\n");
		return NULL;
	}
	return rep;
}

struct authorization_data *
pkinit_decode_authorization_data(PLArenaPool *pool, SECItem *item)
{
	struct authorization_data *adata;

	adata = PORT_ArenaZAlloc(pool, sizeof(*adata));
	if (adata == NULL) {
		pkinit_debug("Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, adata, pkinit_authorization_data_template,
			       item) != SECSuccess) {
		pkinit_debug("Error decoding authorization data!\n");
		return NULL;
	}
	return adata;
}

SECItem *
pkinit_encode_draft_pa_pk_as_req(PLArenaPool *pool,
				 struct draft_pa_pk_as_req *req)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, req,
			       pkinit_draft_pa_pk_as_req_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding draft PA-PK-AS-REQ!\n");
	return NULL;
}

SECItem *
pkinit_encode_pa_pk_as_req(PLArenaPool *pool, struct pa_pk_as_req *req)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, req,
			       pkinit_pa_pk_as_req_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding PA-PK-AS-REQ!\n");
	return NULL;
}

SECItem *
pkinit_encode_draft_pa_pk_as_rep(PLArenaPool *pool,
				 struct draft_pa_pk_as_rep *rep)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, rep,
			       pkinit_draft_pa_pk_as_rep_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding draft PA-PK-AS-REP!\n");
	return NULL;
}

SECItem *
pkinit_encode_pa_pk_as_rep(PLArenaPool *pool, struct pa_pk_as_rep *rep)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, rep,
			       pkinit_pa_pk_as_rep_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding PA-PK-AS-REP!\n");
	return NULL;
}

SECItem *
pkinit_encode_draft_auth_pack(PLArenaPool *pool,
			      struct draft_auth_pack *auth_pack)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, auth_pack,
			       pkinit_draft_auth_pack_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding draft_auth_pack!\n");
	return NULL;
}

SECItem *
pkinit_encode_auth_pack(PLArenaPool *pool,
			struct auth_pack *auth_pack)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, auth_pack,
			       pkinit_auth_pack_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding draft_auth_pack!\n");
	return NULL;
}

SECItem *
pkinit_encode_draft_reply_key_pack(PLArenaPool *pool,
				   struct draft_reply_key_pack *reply_key_pack)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, reply_key_pack,
			       pkinit_draft_reply_key_pack_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding draft_reply_key_pack!\n");
	return NULL;
}

SECItem *
pkinit_encode_reply_key_pack(PLArenaPool *pool,
			     struct reply_key_pack *reply_key_pack)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, reply_key_pack,
			       pkinit_reply_key_pack_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding draft_reply_key_pack!\n");
	return NULL;
}

SECItem *
pkinit_encode_kdc_dh_key_info(PLArenaPool *pool,
			      struct kdc_dh_key_info *info)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, info,
			       pkinit_kdc_dh_key_info_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding KDC DH key info!\n");
	return NULL;
}

SECItem *
pkinit_encode_kerberos_name(PLArenaPool *pool, struct kerberos_name *name)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, name,
			       pkinit_kerberos_name_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding kerberos_name!\n");
	return NULL;
}

SECItem *
pkinit_encode_authorization_data(PLArenaPool *pool,
				 struct authorization_data *adata)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, adata,
			       pkinit_authorization_data_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug("Error encoding authorization_data!\n");
	return NULL;
}

SECItem *
pkinit_create_authorization_data(PLArenaPool *pool,
				 long ad_type,
				 SECItem *ad_data)
{
	struct authorization_data adata;
	memset(&adata, 0, sizeof(adata));
	if (SEC_ASN1EncodeInteger(pool,
				  &adata.ad_type,
				  ad_type) != &adata.ad_type) {
		pkinit_debug("Error encoding authorization data type!\n");
		return NULL;
	}
	if (ad_data != NULL) {
		adata.ad_data = *ad_data;
	}
	return pkinit_encode_authorization_data(pool, &adata);
}

static time_t
get_gmt_offset(void)
{
	struct tm gmtimes, localtimes;
	time_t now, gm, loc;

	now = time(NULL);
	if (now == -1) {
		return 0;
	}

	if (gmtime_r(&now, &gmtimes) == NULL) {
		return 0;
	}
	if (localtime_r(&now, &localtimes) == NULL) {
		return 0;
	}
	loc = mktime(&localtimes);
	if (loc == -1) {
		return 0;
	}
	gm = mktime(&gmtimes);
	if (gm == -1) {
		return 0;
	}
	pkinit_debug("GMT offset is %d.\n", loc - gm);
	return loc - gm;
}

static SECStatus
kerberos_time_from_time(PLArenaPool *pool, time_t *when, SECItem *ktime)
{
	time_t now;
	struct tm tm;

	if (when == NULL) {
		now = time(NULL);
	} else {
		now = *when;
	}

	if (gmtime_r(&now, &tm) != &tm) {
		pkinit_debug("Error parsing current time!\n");
		return SECFailure;
	}

	/* The current time is a GeneralTime, despite the draft docs stating
	 * that it's an integer. */
	ktime->data = PORT_ArenaZAlloc(pool, 16);
	if (ktime->data == NULL) {
		pkinit_debug("Error encoding ctime: out of memory!\n");
		return SECFailure;
	}
	ktime->len = snprintf((char *) ktime->data, 16,
			      "%04d%02d%02d%02d%02d%02dZ",
			      (tm.tm_year + 1900) % 10000,
			      (tm.tm_mon + 1) % 13,
			      (tm.tm_mday % 32),
			      tm.tm_hour, tm.tm_min, tm.tm_sec);
	return SECSuccess;
}

static time_t
time_from_kerberos_time(SECItem *item)
{
	struct tm tm;
	time_t ctime;
	char buf[5];
	memset(&tm, 0, sizeof(tm));

	memset(buf, '\0', sizeof(buf));
	memcpy(buf, item->data + 0, 4);
	tm.tm_year = atoi(buf) - 1900;

	memset(buf, '\0', sizeof(buf));
	memcpy(buf, item->data + 4, 2);
	tm.tm_mon = atoi(buf) - 1;

	memset(buf, '\0', sizeof(buf));
	memcpy(buf, item->data + 6, 2);
	tm.tm_mday = atoi(buf);

	memset(buf, '\0', sizeof(buf));
	memcpy(buf, item->data + 8, 2);
	tm.tm_hour = atoi(buf);

	memset(buf, '\0', sizeof(buf));
	memcpy(buf, item->data + 10, 2);
	tm.tm_min = atoi(buf);

	memset(buf, '\0', sizeof(buf));
	memcpy(buf, item->data + 12, 2);
	tm.tm_sec = atoi(buf);

	ctime = mktime(&tm);
	if (ctime != -1) {
		return ctime + get_gmt_offset();
	} else {
		return 0;
	}
}

/* Verify that the contents of the auth_pack are acceptible. */
static krb5_error_code
pkinit_verify_auth_pack(PLArenaPool *pool,
			krb5_context kcontext,
			krb5_kdc_req *req_body,
			krb5_data *request_packet,
			krb5_deltat max_time_skew,
			struct auth_pack *pack)
{
	SECStatus status;
	struct kdc_req kdc_req;
	struct pkauthenticator *pkauth;
	unsigned char digest[20]; /* FIXME: future-proof this */
	unsigned long nonce;
	SECItem request_packet_item;
	krb5_timestamp now;
	time_t ctime;

	pkauth = &pack->pkauthenticator;

	/* Compare the nonce. */
	status = SEC_ASN1DecodeInteger(&pkauth->nonce, &nonce);
	if (status != SECSuccess) {
		pkinit_debug("Error decoding client nonce!\n");
		return KRB5KRB_ERR_GENERIC;
	}

	if (nonce != (unsigned long) req_body->nonce) {
		pkinit_debug("Nonce mismatch!\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Check that the time is within the acceptible window. */
	if ((pkauth->ctime.len != 15) || (pkauth->ctime.data[14] != 'Z')) {
		pkinit_debug("Error parsing client time!\n");
		return KRB5KRB_ERR_GENERIC;
	}
	ctime = time_from_kerberos_time(&pkauth->ctime);
	if (ctime == 0) {
		pkinit_debug("Error parsing client's time!\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (krb5_timeofday(kcontext, &now) != 0) {
		pkinit_debug("Error determining current time!\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (labs(now - ctime) > max_time_skew) {
		pkinit_debug("Now is %d, client is in %d.\n", now, ctime);
		pkinit_debug("Clock skew error!\n");
		return KRB5KRB_AP_ERR_SKEW;
	}

	/* Take the digest over the KDC-REQ-BODY and compare it to the value
	 * in the auth_pack. */
	request_packet_item.data = (unsigned char *) request_packet->data;
	request_packet_item.len = request_packet->length;
	memset(&kdc_req, 0, sizeof(kdc_req));
	status = SEC_ASN1DecodeItem(pool, &kdc_req,
				    pkinit_kdc_as_req_template,
				    &request_packet_item);

	if (status != SECSuccess) {
		pkinit_debug("Error decoding request packet!\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (pkauth->pa_checksum.len != 20) {
		pkinit_debug("Incorrect client checksum length or "
			     "missing checksum!\n");
		return KRB5KDC_ERR_PA_CHECKSUM_MUST_BE_INCLUDED;
	}
	if (PK11_HashBuf(SEC_OID_SHA1, &digest[0],
			 (unsigned char *) kdc_req.req_body.data,
			 kdc_req.req_body.len) != SECSuccess) {
		pkinit_debug("Error computing client checksum!\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (memcmp(digest, pkauth->pa_checksum.data, 20) != 0) {
		pkinit_debug("Checksum mismatch!\n");
		return KRB5KDC_ERR_DIGEST_IN_SIGNED_DATA_NOT_ACCEPTED;
	}

	pkinit_debug("Checksum matches.\n");
	return 0;
}

/* Build a client public key. */
static SECItem *
pkinit_create_client_public_value(PLArenaPool *pool, int preferred_group,
				  SECKEYPrivateKey **private_key,
				  SECKEYPublicKey **public_key,
				  SECItem **nonce)
{
	struct subject_public_key_info spki;
	struct domain_parameters dh_params;
	SECItem *encoded_params, *encoded_pubval, encoded_info;
	SECOidData *oid;
	PK11SlotInfo *slot;
	SECKEYPrivateKey *priv;
	SECKEYPublicKey *pub;
	SECKEYDHParams params, *oakley;
	PQGParams *pqg_params;
	PQGVerify *pqg_verify;

	/* Initialize the structure. */
	memset(&spki, 0, sizeof(spki));
	*nonce = NULL;

	/* Set the algorithm type. */
	oid = SECOID_FindOIDByTag(SEC_OID_X942_DIFFIE_HELMAN_KEY);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return NULL;
	}

	/* Generate domain parameters. */
	memset(&dh_params, 0, sizeof(dh_params));
	oakley = oakley_get_group(pool, preferred_group);
	if (oakley != NULL) {
		pkinit_debug("Using Oakley group %d.\n", preferred_group);
		params.arena = pool;
		params.prime = oakley->prime;
		params.base = oakley->base;
		dh_params.p = oakley->prime;
		dh_params.g = oakley->base;
		dh_params.q.data = (unsigned char *) "\001";
		dh_params.q.len = 1;
	} else {
		pkinit_debug("Generating new PQG parameters.\n");
		if (PK11_PQG_ParamGen(8, &pqg_params,
				      &pqg_verify) != SECSuccess) {
			pkinit_debug("Error generating parameters.\n");
			return NULL;
		}
		pkinit_debug("Finished generating new PQG parameters.\n");
		params.arena = pool;
		params.prime = pqg_params->prime;
		params.base = pqg_params->base;
		dh_params.p = pqg_params->prime;
		dh_params.g = pqg_params->base;
		dh_params.q = pqg_params->subPrime;
	}

	/* Generate the keying data. */
	slot = PK11_GetBestSlot(CKM_DH_PKCS_KEY_PAIR_GEN, NULL);
	if (slot == NULL) {
		pkinit_debug("Error finding slot for generating DH data.\n");
		return NULL;
	}
	pub = NULL;
	priv = PK11_GenerateKeyPair(slot, CKM_DH_PKCS_KEY_PAIR_GEN,
				    &params, &pub, PR_FALSE, PR_FALSE, NULL);
	if ((priv == NULL) || (pub == NULL)) {
		pkinit_debug("Error generating server DH keying data.\n");
		return NULL;
	}

	/* Encode the parameters. */
	encoded_params = common_encode_domain_parameters(pool, &dh_params);
	if (encoded_params == NULL) {
		pkinit_debug("Error encoding DH parameters.\n");
		return NULL;
	}
	encoded_pubval = common_encode_integer(pool, &pub->u.dh.publicValue);
	if (encoded_pubval == NULL) {
		pkinit_debug("Error encoding DH public value.\n");
		return NULL;
	}

	/* Finish building the SubjectPublicKeyInfo. */
	spki.algorithm.algorithm = oid->oid;
	spki.algorithm.parameters = encoded_params;
	spki.subject_public_key = *encoded_pubval;
	/* The encoder expects "len" to be bits because this is a bit string. */
	spki.subject_public_key.len *= 8;
	if (SEC_ASN1EncodeItem(pool, &encoded_info, &spki,
			       common_subject_public_key_info_template) !=
	    &encoded_info) {
		pkinit_debug("Error encoding subject public key info.\n");
		return NULL;
	}
	*private_key = priv;
	*public_key = pub;

	return SECITEM_ArenaDupItem(pool, &encoded_info);
}

/* Create a new draft_auth_pack. */
static SECItem *
pkinit_create_draft_auth_pack(PLArenaPool *pool,
			      krb5_context kcontext, krb5_principal kdc,
			      krb5_int32 nonce,
			      SECKEYPrivateKey **dh_private_key,
			      SECKEYPublicKey **dh_public_key,
			      SECItem **dh_nonce)
{
	struct draft_auth_pack auth_pack;
	struct draft_pkauthenticator *pk;
	SECItem encoded, *item, *dh_spki;
	krb5_timestamp know;
	krb5_int32 cusec;
	time_t now;
	int i;

	if ((kcontext == NULL) || (kdc == NULL)) {
		pkinit_debug("No key information available!\n");
		return NULL;
	}

	memset(&auth_pack, 0, sizeof(auth_pack));
	pk = &auth_pack.pkauthenticator;

	/* force DH key delivery off */
	dh_private_key = NULL;
	dh_public_key = NULL;

	/* kdc_name (type) */
	if (SEC_ASN1EncodeInteger(pool,
				  &pk->kdc_name.name_type,
				  kdc->type ? kdc->type : 1) !=
	    &pk->kdc_name.name_type) {
		pkinit_debug("Error encoding name type!\n");
		return NULL;
	}
	/* kdc_name (strings) */
	pk->kdc_name.name_string = PORT_ArenaZAlloc(pool,
						    sizeof(SECItem*) *
						    (kdc->length + 1));
	if (pk->kdc_name.name_string == NULL) {
		pkinit_debug("Error allocating memory for name!\n");
		return NULL;
	}
	for (i = 0; i < kdc->length; i++) {
		pk->kdc_name.name_string[i] =
			common_secitem_from_krb5_data(pool, &kdc->data[i]);
		if (pk->kdc_name.name_string[i] == NULL) {
			pkinit_debug("Error allocating memory for "
				     "name component!\n");
			return NULL;
		}
	}
	/* kdc_realm */
	item = common_secitem_from_krb5_data(pool, &kdc->realm);
	if (item == NULL) {
		pkinit_debug("Error allocating memory for realm name!\n");
		return NULL;
	}
	pk->kdc_realm.name = *item;

	/* Get the time of day. */
	if (krb5_us_timeofday(kcontext, &know, &cusec) != 0) {
		pkinit_debug("Error determining current time!\n");
		return NULL;
	}
	now = know;

	/* The current time is a GeneralTime, despite the draft docs stating
	 * that it's an integer. */
	if (kerberos_time_from_time(pool, &now, &pk->ctime) != SECSuccess) {
		pkinit_debug("Error encoding current time!\n");
		return NULL;
	}

	/* The microsecond time is an integer. */
	if (SEC_ASN1EncodeInteger(pool, &pk->cusec, cusec) != &pk->cusec) {
		pkinit_debug("Error encoding cusec!\n");
		return NULL;
	}

	if (SEC_ASN1EncodeInteger(pool, &pk->nonce, nonce) != &pk->nonce) {
		pkinit_debug("Error encoding nonce!\n");
		return NULL;
	}

	/* Add a DH client value. */
	if ((dh_private_key != NULL) && (dh_public_key != NULL)) {
		dh_spki = pkinit_create_client_public_value(pool, 14,
							    dh_private_key,
							    dh_public_key,
							    dh_nonce);
		auth_pack.client_public_value = *dh_spki;
	}

	/* Encode the structure. */
	if (SEC_ASN1EncodeItem(pool, &encoded, &auth_pack,
			       pkinit_draft_auth_pack_template) != &encoded) {
		pkinit_debug("Error encoding draft_auth_pack.\n");
		return NULL;
	}

	return SECITEM_ArenaDupItem(pool, &encoded);
}

/* Create a new auth_pack. */
static SECItem *
pkinit_create_auth_pack(PLArenaPool *pool,
			krb5_context kcontext,
			krb5_kdc_req *request,
			krb5_data *request_body,
			SECKEYPrivateKey **dh_private_key,
			SECKEYPublicKey **dh_public_key,
			SECItem **dh_nonce)
{
	struct auth_pack auth_pack;
	struct pkauthenticator *pk;
	unsigned char digest[20];
	SECItem encoded, *dh_spki;
	krb5_timestamp know;
	krb5_int32 cusec, nonce;
	struct tm tm;
	time_t now;
	SECItem *null_item, rc2_version, rc2_params, *rc2_params_ptr;
	struct {
		SECOidTag tag;
		SECItem **pparams;
	} supported_cms_types[] = {
		{SEC_OID_CMS_3DES_KEY_WRAP, NULL},
		{SEC_OID_CMS_RC2_KEY_WRAP, &rc2_params_ptr},
		{SEC_OID_SHA1, NULL},
		{SEC_OID_MD5, &null_item},
		{SEC_OID_PKCS1_RSA_ENCRYPTION, &null_item},
	};
	struct algorithm_identifier *algorithm, **supported_cms_type_list;
	unsigned int i, j;
	size_t size;
	SECOidData *oid;
	SECKEYPrivateKey *priv;
	SECKEYPublicKey *pub;

	memset(&auth_pack, 0, sizeof(auth_pack));
	*dh_private_key = NULL;
	*dh_public_key = NULL;
	*dh_nonce = NULL;

	/* Get the time of day. */
	if (krb5_us_timeofday(kcontext, &know, &cusec) != 0) {
		pkinit_debug("Error determining current time!\n");
		return NULL;
	}
	now = know;
	if (gmtime_r(&now, &tm) != &tm) {
		pkinit_debug("Error parsing current time!\n");
		return NULL;
	}

	/* The current time is a GeneralTime, despite the draft docs claiming
	 * that it's an integer. */
	pk = &auth_pack.pkauthenticator;
	pk->ctime.data = PORT_ArenaZAlloc(pool, 16);
	if (pk->ctime.data == NULL) {
		pkinit_debug("Error encoding ctime: out of memory!\n");
		return NULL;
	}
	pk->ctime.len = snprintf((char *) pk->ctime.data, 16,
				 "%04d%02d%02d%02d%02d%02dZ",
				 (tm.tm_year + 1900) % 10000,
				 (tm.tm_mon + 1) % 13,
				 (tm.tm_mday % 32),
				 tm.tm_hour, tm.tm_min, tm.tm_sec);

	/* The microsecond time is an integer. */
	if (SEC_ASN1EncodeInteger(pool, &pk->cusec, cusec) != &pk->cusec) {
		pkinit_debug("Error encoding cusec!\n");
		return NULL;
	}

	/* The spec says that this nonce doesn't have to be the same as the
	 * request nonce.  But for now, at least, just use that value. FIXME */
	nonce = request->nonce;
	if (SEC_ASN1EncodeInteger(pool, &pk->nonce, nonce) != &pk->nonce) {
		pkinit_debug("Error encoding nonce!\n");
		return NULL;
	}

	/* Generate the checksum of the encoded request body. */
	if (PK11_HashBuf(SEC_OID_SHA1, &digest[0],
			 (unsigned char *) request_body->data,
			 request_body->length) == SECSuccess) {
		pk->pa_checksum.len = 20;
		pk->pa_checksum.data = digest;
	} else {
		pk->pa_checksum.len = 0;
		pk->pa_checksum.data = NULL;
	}

	/* This is used if we're using DH. */
	priv = NULL;
	pub = NULL;
	if ((dh_public_key != NULL) && (dh_private_key != NULL)) {
		dh_spki = pkinit_create_client_public_value(pool, 14, &priv,
							    &pub, dh_nonce);
		auth_pack.client_public_value = *dh_spki;
	}
	if ((auth_pack.client_public_value.len > 0) && (dh_nonce != NULL)) {
		*dh_nonce = common_make_random_item(pool, kcontext, 256 / 8);
		if (*dh_nonce == NULL) {
			pkinit_debug("Error generating client DH nonce!\n");
			return NULL;
		}
		auth_pack.client_dh_nonce = **dh_nonce;
	}

	/* Create a list of supported CMS algorithms. */
	size = sizeof(supported_cms_types) / sizeof(supported_cms_types[0]) + 1;
	size = size * sizeof(struct algorithm_identifier *);
	supported_cms_type_list = PORT_ArenaZAlloc(pool, size);
	if (supported_cms_type_list == NULL) {
		pkinit_debug("Out of memory!\n");
		return NULL;
	}

	/* Create a NULL item. */
	null_item = &pkinit_null_item;

	/* Create the RC2 parameters (version). */
	if (SEC_ASN1EncodeInteger(pool, &rc2_version, 0x3a) != &rc2_version) {
		pkinit_debug("Error encoding RC2 version!\n");
		return NULL;
	}
	rc2_params_ptr = SEC_ASN1EncodeItem(pool, &rc2_params, 
					    &rc2_version, SEC_IntegerTemplate);

	/* Populate the CMS type list, with parameters. */
	for (i = 0, j = 0;
	     i < sizeof(supported_cms_types) /
	         sizeof(supported_cms_types[0]);
	     i++) {
		oid = SECOID_FindOIDByTag(supported_cms_types[i].tag);
		if (oid != NULL) {
			size = sizeof(struct algorithm_identifier);
			algorithm = PORT_ArenaZAlloc(pool, size);
			if (algorithm != NULL) {
				algorithm->algorithm = oid->oid;
				if (supported_cms_types[i].pparams) {
					algorithm->parameters =
						*supported_cms_types[i].pparams;
				} else {
					algorithm->parameters = NULL;
				}
				supported_cms_type_list[j++] = algorithm;
			}
		}
	}
	auth_pack.supported_cms_types = supported_cms_type_list;

	/* Encode the structure. */
	if (SEC_ASN1EncodeItem(pool, &encoded, &auth_pack,
			       pkinit_auth_pack_template) != &encoded) {
		pkinit_debug("Error encoding auth_pack.\n");
		return NULL;
	}

	if (dh_private_key != NULL) {
		*dh_private_key = priv;
	}
	if (dh_public_key != NULL) {
		*dh_public_key = pub;
	}

	return SECITEM_ArenaDupItem(pool, &encoded);
}

/* Given the name of a member of the realm, construct the name of the
 * ticket-granting service. */
krb5_error_code
pkinit_build_tgs_name(krb5_context kcontext, krb5_principal member,
		      krb5_principal *tgs_principal_name)
{
	*tgs_principal_name = NULL;
	return krb5_build_principal_ext(kcontext, tgs_principal_name,
					member->realm.length,
					member->realm.data,
					KRB5_TGS_NAME_SIZE,
					KRB5_TGS_NAME,
					member->realm.length,
					member->realm.data,
					0);
}

/* Build a new PA-PK-AS-REQ (draft version). */
krb5_data *
pkinit_create_draft_pa_pk_as_req(PLArenaPool *pool,
				 krb5_context kcontext,
				 krb5_kdc_req *request,
				 CERTCertDBHandle *certdb,
				 CERTCertificate *cert,
				 PK11SlotInfo *preferred_slot,
				 SECKEYPrivateKey **dh_private_key,
				 SECKEYPublicKey **dh_public_key,
				 SECItem **dh_nonce,
				 struct pkinit_pwcb_args *pwcb_args)
{
	SECItem *auth_pack, *signed_auth_pack;
	SECOidData *oid;
	struct draft_pa_pk_as_req req;
	krb5_principal tgt_princ;
	SECItem *ret;

	memset(&req, 0, sizeof(req));

	/* Find the KDC's name. */
	tgt_princ = NULL;
	if (pkinit_build_tgs_name(kcontext, request->server, &tgt_princ) != 0) {
		pkinit_debug("Error building KDC principal name!\n");
		return NULL;
	}

	/* Build the auth_pack. */
	auth_pack = pkinit_create_draft_auth_pack(pool,
						  kcontext,
						  tgt_princ,
						  request->nonce,
						  dh_private_key,
						  dh_public_key,
						  dh_nonce);
	if (auth_pack == NULL) {
		pkinit_debug("Internal error creating draft_auth_pack.\n");
		krb5_free_principal(kcontext, tgt_princ);
		return NULL;
	}

	krb5_free_principal(kcontext, tgt_princ);

	/* Sign the auth_pack. */
	signed_auth_pack = bcms_create_signed_data(pool,
						   certdb,
						   NULL,
						   auth_pack,
						   cert,
						   SEC_OID_PKCS1_MD5_WITH_RSA_ENCRYPTION,
						   preferred_slot,
						   pwcb_args);
	if (signed_auth_pack == NULL) {
		pkinit_debug("Error signing draft_auth_pack.\n");
		return NULL;
	}

	/* Fill in the rest of the request. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return NULL;
	}
	req.signed_auth_pack.content_type = oid->oid;
	req.signed_auth_pack.content = *signed_auth_pack;

	/* For now, leave the list of trusted CAs and the KDC's cert empty. */
	/* FIXME */

	ret = pkinit_encode_draft_pa_pk_as_req(pool, &req);
	return common_krb5_data_from_secitem(kcontext, ret);
}

/* Build a new PA-PK-AS-REQ. */
krb5_data *
pkinit_create_pa_pk_as_req(PLArenaPool *pool,
			   krb5_context kcontext,
			   krb5_kdc_req *request,
			   krb5_data *encoded_request_body,
			   CERTCertDBHandle *certdb,
			   CERTCertificate *cert,
			   SECOidTag digest,
			   PK11SlotInfo *preferred_slot,
			   SECKEYPrivateKey **dh_priv,
			   SECKEYPublicKey **dh_pub,
			   SECItem **dh_nonce,
			   struct pkinit_pwcb_args *pwcb_args)
{
	SECItem *auth_pack, *signed_auth_pack, *oid_wdata;
	SECOidData *oid;
	const SECItem *oid_data;
	struct pa_pk_as_req req;
	SECItem *ret;

	memset(&req, 0, sizeof(req));

	/* Find the OID for the signed data content. */
	oid_data = cert_get_oid_pkinit_auth_data();
	if (oid_data == NULL) {
		pkinit_debug("Internal error.\n");
		return NULL;
	}
	oid_wdata = SECITEM_ArenaDupItem(pool, oid_data);

	/* Build the auth_pack. */
	auth_pack = pkinit_create_auth_pack(pool, kcontext, request,
					    encoded_request_body,
					    dh_priv, dh_pub, dh_nonce);
	if (auth_pack == NULL) {
		pkinit_debug("Internal error creating auth_pack.\n");
		return NULL;
	}

	/* Sign the auth_pack. */
	signed_auth_pack = bcms_create_signed_data(pool,
						   certdb,
						   oid_wdata,
						   auth_pack,
						   cert,
						   digest,
						   preferred_slot,
						   pwcb_args);
	if (signed_auth_pack == NULL) {
		pkinit_debug("Error signing draft_auth_pack.\n");
		return NULL;
	}

	/* Fill in the rest of the request. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return NULL;
	}
	req.signed_auth_pack.content_type = oid->oid;
	req.signed_auth_pack.content = *signed_auth_pack;

	/* For now, leave the list of trusted CAs and the KDC's cert empty. */
	/* FIXME? */

	ret = pkinit_encode_pa_pk_as_req(pool, &req);
	return common_krb5_data_from_secitem(kcontext, ret);
}

/* Create a key using a given octet string as keying material. */
static SECStatus
pkinit_octet_string_to_aeskey(krb5_context kcontext,
			      SECOidTag hash_alg,
			      SECItem *input, SECItem *input2, SECItem *input3,
			      krb5_enctype enctype, krb5_keyblock *keyblock)
{
	PK11Context *ctx;
	size_t left;
	unsigned int length;
	unsigned char buf[512 / 8];
	unsigned char c;
	unsigned char key_buf[256 / 8];
	krb5_keyblock key;

	switch (hash_alg) {
	case SEC_OID_MD4:
	case SEC_OID_MD5:
	case SEC_OID_SHA1:
	case SEC_OID_SHA256:
	case SEC_OID_SHA384:
	case SEC_OID_SHA512:
		break;
	default:
		pkinit_debug("Unsupported hash deriving key!\n");
		return SECFailure;
	}

	switch (enctype) {
	case ENCTYPE_AES128_CTS_HMAC_SHA1_96:
		left = 128 / 8;
		break;
	case ENCTYPE_AES256_CTS_HMAC_SHA1_96:
		left = 256 / 8;
		break;
	default:
		pkinit_debug("Unsupported enctype deriving key!\n");
		return SECFailure;
	}

	c = 0;
	key.contents = key_buf;
	key.length = left;
	key.enctype = enctype;

	while (left > 0) {
		ctx = PK11_CreateDigestContext(hash_alg);
		if (ctx == NULL) {
			pkinit_debug("Error setting up hash!\n");
			return SECFailure;
		}
		if (PK11_DigestOp(ctx, &c, 1) != SECSuccess) {
			pkinit_debug("Error digesting counter!\n");
			return SECFailure;
		}
		pkinit_debug("Digesting %d-byte shared secret to derive key, "
			     "round %d.\n", input->len, c + 1);
		if (PK11_DigestOp(ctx,
				  (unsigned char *) input->data,
				  input->len) != SECSuccess) {
			pkinit_debug("Error digesting shared secret!\n");
			return SECFailure;
		}
		if ((input2 != NULL) && (input2->len > 0)) {
			if (PK11_DigestOp(ctx,
					  (unsigned char *) input2->data,
					  input2->len) != SECSuccess) {
				pkinit_debug("Error digesting nonce!\n");
				return SECFailure;
			}
		}
		if ((input3 != NULL) && (input3->len > 0)) {
			if (PK11_DigestOp(ctx,
					  (unsigned char *) input3->data,
					  input3->len) != SECSuccess) {
				pkinit_debug("Error digesting nonce!\n");
				return SECFailure;
			}
		}
		if (PK11_DigestFinal(ctx, buf, &length,
				     sizeof(buf)) != SECSuccess) {
			pkinit_debug("Error digesting shared data + nonce!\n");
			return SECFailure;
		}
		if (left < length) {
			length = left;
		}
		memcpy(key.contents + key.length - left, buf, length);
		left -= length;
		c++;
	}

	if (krb5_copy_keyblock_contents(kcontext, &key, keyblock) != 0) {
		pkinit_debug("Error copying keyblock!\n");
		return SECFailure;
	}
	return SECSuccess;
}

/* Create a new reply_key_pack. */
static SECItem *
pkinit_create_reply_key_pack(PLArenaPool *pool,
			     krb5_context kcontext,
			     krb5_keyblock *encrypting_key,
			     krb5_data *encoded_request)
{
	struct reply_key_pack reply_key_pack;
	krb5_keyblock reply_key;
	krb5_checksum checksum;

	if (kcontext == NULL) {
		pkinit_debug("No key information available!\n");
		return NULL;
	}
	memset(&reply_key_pack, 0, sizeof(reply_key_pack));

	/* Generate a new encryption key. */
	memset(&reply_key, 0, sizeof(reply_key));
	reply_key.length = 128 / 8;
	if (krb5_c_make_random_key(kcontext, ENCTYPE_AES256_CTS_HMAC_SHA1_96,
				   &reply_key) != 0) {
		pkinit_debug("Error generating temporary AS reply key!\n");
		return NULL;
	}
	if (krb5_copy_keyblock_contents(kcontext, &reply_key,
					encrypting_key) != 0) {
		pkinit_debug("Error saving temporary AS reply key!\n");
		return NULL;
	}

	/* Generate the checksum of the request packet (the entire AS-REQ). */
	memset(&checksum, 0, sizeof(checksum));
	if (krb5_c_make_checksum(kcontext, CKSUMTYPE_HMAC_SHA1_DES3, &reply_key,
				 6, encoded_request, &checksum) != 0) {
		pkinit_debug("Internal error generating keyed checksum!\n");
		krb5_free_keyblock_contents(kcontext, &reply_key);
		return NULL;
	}

	/* Copy the key to the reply_key_pack. */
	if (SEC_ASN1EncodeInteger(pool, &reply_key_pack.reply_key.keytype,
				  reply_key.enctype) !=
	    &reply_key_pack.reply_key.keytype) {
		pkinit_debug("Internal error encoding key type!\n");
		krb5_free_keyblock_contents(kcontext, &reply_key);
		return NULL;
	}
	reply_key_pack.reply_key.keyvalue.len = reply_key.length;
	reply_key_pack.reply_key.keyvalue.data =
		PORT_ArenaZAlloc(pool, reply_key.length);
	if (reply_key_pack.reply_key.keyvalue.data == NULL) {
		pkinit_debug("Out of memory!\n");
		krb5_free_keyblock_contents(kcontext, &reply_key);
		return NULL;
	}
	memcpy(reply_key_pack.reply_key.keyvalue.data,
	       reply_key.contents, reply_key.length);
	krb5_free_keyblock_contents(kcontext, &reply_key);

	/* Copy the checksum to the reply_key_pack. */
	if (SEC_ASN1EncodeInteger(pool, &reply_key_pack.as_checksum.cksumtype,
				  checksum.checksum_type) !=
	    &reply_key_pack.as_checksum.cksumtype) {
		pkinit_debug("Internal error encoding keyed checksum type!\n");
		return NULL;
	}
	reply_key_pack.as_checksum.checksum.data =
		PORT_ArenaAlloc(pool, checksum.length);
	if (reply_key_pack.as_checksum.checksum.data == NULL) {
		pkinit_debug("Internal error encoding keyed checksum!\n");
		return NULL;
	}
	memcpy(reply_key_pack.as_checksum.checksum.data,
	       checksum.contents, checksum.length);
	reply_key_pack.as_checksum.checksum.len = checksum.length;
			
	return pkinit_encode_reply_key_pack(pool, &reply_key_pack);
}

/* Verify a PA-PK-AS-REQ structure. */
krb5_error_code
pkinit_verify_pa_pk_as_req(PLArenaPool *pool,
			   CERTCertDBHandle *certdb,
			   krb5_context kcontext,
			   krb5_kdc_req *request,
			   krb5_data *request_pkt,
			   krb5_data *padata,
			   krb5_deltat max_time_skew,
			   SECItem **authorized_certs,
			   int n_authorized_certs,
			   CERTCertificate **signing_cert,
			   struct subject_public_key_info **client_dh_key,
			   SECItem **nonce,
			   SECItem **client_dh_nonce,
			   krb5_authdata **adata)
{
	struct pa_pk_as_req *req;
	struct signed_data *signed_data;
	struct subject_public_key_info *spki;
	struct external_principal_identifier **verified_ca;
	struct auth_pack *auth_pack;
	SECItem *padata_item, *auth_data, encoded_cas;
	const SECItem *roid_data;
	SECOidData *oid;
	CERTCertificate *signer, *issuer;
	SECStatus status;
	krb5_error_code ret;
	int mapped, i, n_verified_ca;

	*signing_cert = NULL;
	*adata = NULL;

	/* Parse the data (outer level). */
	padata_item = common_secitem_from_krb5_data(pool, padata);
	req = pkinit_decode_pa_pk_as_req(pool, padata_item);
	if (req == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Check that we know how to handle it. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (!SECITEM_ItemsAreEqual(&oid->oid,
				   &req->signed_auth_pack.content_type)) {
		pkinit_debug("Data is not signed data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Decode the signed data. */
	signed_data = bcms_decode_signed_data(pool,
					      &req->signed_auth_pack.content);
	if (signed_data == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Verify the signature. */
	signer = NULL;
	status = bcms_verify_signed_data(certdb, signed_data, &signer);
	if (status != SECSuccess) {
		pkinit_debug("Verify error.\n");
		return KRB5KDC_ERR_INVALID_SIG;
	}

	/* Validate the certificate. */
	if (signer == NULL) {
		pkinit_debug("No signer?\n");
		return KRB5KDC_ERR_CANT_VERIFY_CERTIFICATE;
	}

	if (CERT_VerifyCertificateNow(certdb, signer, PR_TRUE,
				      certificateUsageSSLClient,
				      NULL, NULL) != SECSuccess) {
		pkinit_debug("Client certificate failed validation!\n");
		return KRB5KDC_ERR_CLIENT_NOT_TRUSTED;
	}

	/* Check if the cert belongs to the user. */
	mapped = 0;
	if (authorized_certs != NULL) {
		for (i = 0; i < n_authorized_certs; i++) {
			if (SECITEM_ItemsAreEqual(authorized_certs[i],
						  &signer->derCert)) {
				mapped++;
			}
		}
	}
	if (!mapped) {
		/* Check if the cert is acceptable using the usual means. */
		ret = cert_validate_client_certificate(certdb,
						       signer,
						       NULL,
						       kcontext,
						       request->client);
		if (ret != 0) {
			pkinit_debug("Client certificate not acceptible.\n");
			return ret;
		}
	}

	/* Check that the signed data is what we're expecting. */
	roid_data = cert_get_oid_pkinit_auth_data();
	if (!SECITEM_ItemsAreEqual(roid_data,
				   &signed_data->encap_content_info.content_type)) {
		pkinit_debug("Signed data is not PKINIT data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Decode the auth_pack. */
	auth_pack = pkinit_decode_auth_pack(pool,
					    &signed_data->encap_content_info.content);
	if (auth_pack == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Verify the auth_pack. */
	ret = pkinit_verify_auth_pack(pool,
				      kcontext,
				      request,
				      request_pkt,
				      max_time_skew,
				      auth_pack);
	if (ret != 0) {
		pkinit_debug("Error in auth_pack.\n");
		return ret;
	}

	/* Save the client's certificate. */
	if (signing_cert != NULL) {
		*signing_cert = signer;
	}
	/* Save the client's nonce. */
	if (nonce != NULL) {
		*nonce = &auth_pack->pkauthenticator.nonce;
	}
	/* Save the client's DH nonce, if one was provided. */
	if (client_dh_nonce != NULL) {
		*client_dh_nonce = &auth_pack->client_dh_nonce;
	}
	/* Save the client's DH keying information, if some was provided. */
	if (client_dh_key != NULL) {
		if (auth_pack->client_public_value.len > 0) {
			spki = common_decode_subject_public_key_info(pool,
								     &auth_pack->client_public_value);
			if (spki == NULL) {
				pkinit_debug("Error decoding subject public key info!\n");
			} else {
				*client_dh_key = spki;
				pkinit_debug("Saved subject public key info.\n");
			}
		}
	}
	/* Build the list of validated certificates. */
	issuer = CERT_FindCertIssuer(signer, 0, 0);
	n_verified_ca = 0;
	while ((issuer != NULL) && !issuer->isRoot) {
		n_verified_ca++;
		issuer = CERT_FindCertIssuer(issuer, 0, 0);
	}
	if (n_verified_ca > 0) {
		verified_ca = PORT_ArenaZAlloc(pool,
					       sizeof(verified_ca[0]) *
					       (n_verified_ca + 1));
		if (verified_ca != NULL) {
			issuer = CERT_FindCertIssuer(signer, 0, 0);
			i = 0;
			while ((issuer != NULL) && !issuer->isRoot) {
				verified_ca[i] = PORT_ArenaZAlloc(pool,
								  sizeof(*(verified_ca[i])));
				if (verified_ca[i] != NULL) {
					verified_ca[i]->der_subject_name = issuer->derSubject;
					if (issuer->subjectKeyID.len > 0) {
						verified_ca[i]->subject_key_identifier = issuer->subjectKeyID;
					}
				}
				issuer = CERT_FindCertIssuer(issuer, 0, 0);
				i++;
			}
		}
		auth_data = NULL;
		if ((SEC_ASN1EncodeItem(pool, &encoded_cas, &verified_ca,
				        bcms_sequence_of_external_principal_identifiers_template) == &encoded_cas) &&
		    (encoded_cas.len > 0)) {
			auth_data = &encoded_cas;
		}
		if (auth_data != NULL) {
			auth_data = pkinit_create_authorization_data(pool,
								     9,
								     auth_data);
		}
		if (auth_data != NULL) {
			*adata = malloc(sizeof(krb5_authdata));
			if (*adata != NULL) {
				(*adata)->contents = malloc(auth_data->len);
				if ((*adata)->contents != NULL) {
					(*adata)->length = auth_data->len;
					memcpy((*adata)->contents,
					       auth_data->data,
					       auth_data->len);
					(*adata)->ad_type = 1;
				}
			}
		}
	}

	/* Success! */
	return 0;
}

/* Given a server and client cert, build the main part of an enc-key-pack
 * reply. */
static krb5_error_code
pkinit_build_reply_key_pack(PLArenaPool *pool,
			    CERTCertDBHandle *certdb,
			    CERTCertificate *server_cert,
			    SECItem *client_cert,
			    krb5_context kcontext,
			    krb5_keyblock *encrypting_key,
			    krb5_data *encoded_request,
			    struct pa_pk_as_rep *rep)
{
	SECItem *reply_key_pack, *signed_reply_key_pack, *enc_key_pack;
	SECOidData *oid;
	const SECItem *signed_content_oid;

	/* Determine the OID for the reply data. */
	signed_content_oid = cert_get_oid_pkinit_rkey_data();
	if (signed_content_oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Build the reply_key_pack. */
	reply_key_pack = pkinit_create_reply_key_pack(pool, kcontext,
						      encrypting_key,
						      encoded_request);
	if (reply_key_pack == NULL) {
		pkinit_debug("Internal error generating reply_key_pack.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Sign the reply_key_pack. */
	signed_reply_key_pack = bcms_create_signed_data(pool,
							certdb,
							signed_content_oid,
							reply_key_pack,
							server_cert,
							SEC_OID_SHA1,
							NULL,
							NULL);
	if (signed_reply_key_pack == NULL) {
		pkinit_debug("Internal error signing reply_key_pack.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Encrypt the signed reply_key_pack for the client. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	enc_key_pack = bcms_create_enveloped_data(pool,
						  certdb,
						  kcontext,
						  &oid->oid,
						  SEC_OID_DES_EDE3_CBC,
						  signed_reply_key_pack,
						  client_cert);
	if (enc_key_pack == NULL) {
		pkinit_debug("Internal error encrypting reply_key_pack.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Build the reply structure. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_ENVELOPED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	memset(rep, 0, sizeof(*rep));
	rep->pa_pk_as_rep_type = pa_pk_as_rep_type_enc_key_pack;
	rep->pa_pk_as_rep_value.enc_key_pack.content_type = oid->oid;
	rep->pa_pk_as_rep_value.enc_key_pack.content = *enc_key_pack;

	/* Success! */
	return 0;
}

/* Given a server and client keying info, build the main part of an
 * dh-key-pack reply. */
static krb5_error_code
pkinit_build_dh_key_info(PLArenaPool *pool,
			 CERTCertDBHandle *certdb,
			 CERTCertificate *server_cert,
			 struct subject_public_key_info *client_key_info,
			 SECItem *client_dh_nonce,
			 krb5_context kcontext,
			 krb5_keyblock *encrypting_key,
			 struct pa_pk_as_rep *rep)
{
	size_t nonce_len;
	struct kdc_dh_key_info kdc_dh_key_info;
	struct domain_parameters *dh_params;
	const SECItem *signed_content_oid;
	SECOidData *oid;
	SECItem *shared_secret, empty_item, *server_dh_nonce, *encoded_pubval;
	SECItem *encoded_kdc_dh_key_info, *signed_data, *client_params;
	SECItem *client_pubval, spk_bits;
	SECKEYDHParams params;
	SECKEYPrivateKey *priv;
	SECKEYPublicKey *pub, client_pub;
	PK11SlotInfo *slot;
	PK11SymKey *sym;

	pkinit_debug("Using DH key agreement.\n");

	memset(&empty_item, 0, sizeof(empty_item));

	/* Disable DH key re-use until we can cache factor/key sets. */
	client_dh_nonce = NULL;

	/* Decode the client's DH parameters. */
	client_params = client_key_info->algorithm.parameters;
	if (client_params == NULL) {
		pkinit_debug("Error retrieving client DH parameters.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	dh_params = common_decode_domain_parameters(pool, client_params);
	if (dh_params == NULL) {
		pkinit_debug("Error decoding client DH parameters.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Generate a key of our own using the client-supplied parameters. */
	slot = PK11_GetBestSlot(CKM_DH_PKCS_KEY_PAIR_GEN, NULL);
	if (slot == NULL) {
		pkinit_debug("Error finding slot for generating DH data.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	memset(&params, 0, sizeof(params));
	params.arena = pool;
	params.prime = dh_params->p;
	params.base = dh_params->g;
	pub = NULL;
	priv = PK11_GenerateKeyPair(slot, CKM_DH_PKCS_KEY_PAIR_GEN,
				    &params, &pub, PR_FALSE, PR_FALSE, NULL);
	if (priv == NULL) {
		pkinit_debug("Error generating server DH keying data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Pull out the client's public value. */
	spk_bits = client_key_info->subject_public_key;
	/* The subject_public_key is a bit string, so the length is a bit
	 * count.  Convert that back to a byte count. */
	spk_bits.len = (spk_bits.len + 7) / 8;
	client_pubval = common_decode_integer(pool, &spk_bits);
	if (client_pubval == NULL) {
		pkinit_debug("Error decoding client public value.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Derive the shared key. */
	client_pub = *pub;
	client_pub.u.dh.publicValue = *client_pubval;
	sym = PK11_PubDerive(priv, &client_pub, PR_FALSE,
			     NULL,
			     NULL,
			     CKM_DH_PKCS_DERIVE,
			     CKM_TLS_MASTER_KEY_DERIVE_DH,
			     CKA_DERIVE,
			     0,
			     NULL);
	if (sym == NULL) {
		pkinit_debug("Error deriving DH shared secret.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Extract the shared secret. */
	if (PK11_ExtractKeyValue(sym) != SECSuccess) {
		pkinit_debug("Can't extract DH shared secret.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	shared_secret = PK11_GetKeyData(sym);
	if (shared_secret == NULL) {
		pkinit_debug("Error extracting DH shared secret.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* If the client supplied a nonce, create a server nonce. */
	if (client_dh_nonce == NULL) {
		server_dh_nonce = NULL;
	} else {
		nonce_len = client_dh_nonce->len;
		if (nonce_len < 256 / 8) {
			nonce_len = 256 / 8;
		}
		server_dh_nonce = common_make_random_item(pool, kcontext,
							  nonce_len);
		if (server_dh_nonce == NULL) {
			pkinit_debug("Error generating server DH nonce.\n");
			return KRB5KRB_ERR_GENERIC;
		}
	}

	/* Derive an encryption key from the shared secret and nonces. */
	if (pkinit_octet_string_to_aeskey(kcontext, SEC_OID_SHA1,
					  shared_secret,
					  client_dh_nonce,
					  server_dh_nonce,
					  ENCTYPE_AES256_CTS_HMAC_SHA1_96,
					  encrypting_key) != SECSuccess) {
		pkinit_debug("Error deriving DH-based encryption key.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Encode the kdc_dh_key_info. */
	memset(&kdc_dh_key_info, 0, sizeof(kdc_dh_key_info));
	encoded_pubval = common_encode_integer(pool, &pub->u.dh.publicValue);
	if (encoded_pubval == NULL) {
		pkinit_debug("Error encoding DH public value.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	kdc_dh_key_info.subject_public_key = *encoded_pubval;
	/* The encoder expects "len" to be bits because this is a bit string. */
	kdc_dh_key_info.subject_public_key.len *= 8;
	kdc_dh_key_info.nonce = client_dh_nonce ?
				*server_dh_nonce :
				empty_item;
	encoded_kdc_dh_key_info = pkinit_encode_kdc_dh_key_info(pool,
								&kdc_dh_key_info);
	if (encoded_kdc_dh_key_info == NULL) {
		pkinit_debug("Error encoding kdc_dh_key_info.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Determine the OID for the reply data. */
	signed_content_oid = cert_get_oid_pkinit_dhkey_data();
	if (signed_content_oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Sign the kdc_dh_key_info. */
	signed_data = bcms_create_signed_data(pool,
					      certdb,
					      signed_content_oid,
					      encoded_kdc_dh_key_info,
					      server_cert,
					      SEC_OID_SHA1,
					      NULL,
					      NULL);

	/* Build the reply structure. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	memset(rep, 0, sizeof(*rep));
	rep->pa_pk_as_rep_type = pa_pk_as_rep_type_dh_info;
	rep->pa_pk_as_rep_value.dh_info.dh_signed_data.content_type = oid->oid;
	rep->pa_pk_as_rep_value.dh_info.dh_signed_data.content = *signed_data;
	rep->pa_pk_as_rep_value.dh_info.server_dh_nonce = client_dh_nonce ?
							  server_dh_nonce :
							  NULL;

	/* Success! */
	return 0;
}

/* Build a new PA-PK-AS-REP. */
krb5_data *
pkinit_create_pa_pk_as_rep(PLArenaPool *pool,
			   CERTCertDBHandle *certdb,
			   krb5_context kcontext,
			   krb5_kdc_req *request,
			   krb5_keyblock *encrypting_key,
			   krb5_data *encoded_request,
			   SECItem *client_cert,
			   struct subject_public_key_info *client_public_key,
			   SECItem *client_dh_nonce)
{
	krb5_principal tgt_princ;
	struct pa_pk_as_rep rep;
	CERTCertificate *cert;
	SECItem *ret;

	/* Find the KDC's certificate. */
	tgt_princ = NULL;
	if (pkinit_build_tgs_name(kcontext, request->server, &tgt_princ) != 0) {
		pkinit_debug("Error building KDC principal name!\n");
		return NULL;
	}
	cert = cert_find_preferred_cert(NULL,
					NULL,
					kcontext,
					tgt_princ,
					CERT_MATCHES_KRB_SAN |
					CERT_MATCHES_UPN_SAN);
	krb5_free_principal(kcontext, tgt_princ);
	if (cert == NULL) {
		pkinit_debug("Unable to locate KDC certificate!\n");
		return NULL;
	}

	/* Build the body of the reply try to use DH if we can. */
	if ((client_public_key == NULL) ||
	    (pkinit_build_dh_key_info(pool, certdb, cert, client_public_key,
				      client_dh_nonce,
				      kcontext,
				      encrypting_key,
				      &rep) != 0)) {
		/* Build an enc-key-pack reply. */
		if (pkinit_build_reply_key_pack(pool, certdb, cert, client_cert,
						kcontext,
						encrypting_key,
						encoded_request,
						&rep) != 0) {
			pkinit_debug("Error building enc_key_pack!\n");
			return NULL;
		}
	}

	/* Finish up. */
	ret = pkinit_encode_pa_pk_as_rep(pool, &rep);
	return common_krb5_data_from_secitem(kcontext, ret);
}

/* Validate the certificate and check that it's an acceptible KDC certificate.
 * for this request. */
static krb5_error_code
pkinit_validate_kdc_certificate(CERTCertDBHandle *certdb,
				CERTCertificate *signer,
				krb5_context kcontext,
				krb5_kdc_req *request)
{
	krb5_principal tgt_princ;
	krb5_error_code ret;

	/* Do the normal verification stuff. */
	if (signer == NULL) {
		pkinit_debug("No signer?\n");
		return KRB5KDC_ERR_CANT_VERIFY_CERTIFICATE;
	}
	if (CERT_VerifyCertificateNow(certdb, signer, PR_TRUE,
				      0,
				      NULL, NULL) != SECSuccess) {
		pkinit_debug("KDC certificate failed validation!\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Do some sanity checks on the server's certificate. */
	tgt_princ = NULL;
	if (pkinit_build_tgs_name(kcontext, request->server, &tgt_princ) != 0) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	ret = cert_validate_kdc_certificate(certdb,
					    signer,
					    NULL,
					    kcontext,
					    tgt_princ);
	if (ret != 0) {
		pkinit_debug("KDC certificate not acceptible.\n");
		krb5_free_principal(kcontext, tgt_princ);
		return ret;
	}
	krb5_free_principal(kcontext, tgt_princ);

	return 0;
}

/* Verify a the enc-key-pack in a PA-PK-AS-REP structure, and pull out the
 * interesting bits. */
static krb5_error_code
pkinit_verify_enc_key_pack_common(PLArenaPool *pool,
				  CERTCertDBHandle *certdb,
				  PK11SlotInfo *preferred_slot,
				  struct pkinit_pwcb_args *pwcb_args,
				  krb5_context kcontext,
				  krb5_kdc_req *request,
				  SECItem *envelope,
				  const SECItem *expected_inner_type,
				  SECItem **inner_data)
{
	struct signed_data *signed_data;
	struct enveloped_data *enveloped_data;
	struct content_info *content_info;
	SECItem *data;
	SECOidData *oid;
	CERTCertificate *signer;
	SECStatus status;
	krb5_error_code ret;

	/* Decode the enveloped data. */
	enveloped_data = bcms_decode_enveloped_data(pool, envelope);
	if (enveloped_data == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	data = bcms_extract_enveloped_data(pool, certdb,
					   enveloped_data,
					   preferred_slot,
					   pwcb_args);
	if (data == NULL) {
		pkinit_debug("Error decrypting server response.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Check the type of the encapsulated data.  Accept either encapsulated
	 * content type = signed data with a signed-data payload or
	 * encapsulated content type = data with a content-type payload which
	 * contains a signed-data. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (SECITEM_ItemsAreEqual(&oid->oid,
				  &enveloped_data->encrypted_content_info.content_type)) {
		/* Try to parse the data as a content-info structure. */
		content_info = bcms_decode_content_info(pool, data);
		if (content_info == NULL) {
			/* Just hope the server's messing with us, and it's
			 * signed data. */
			oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
			content_info = PORT_ArenaZAlloc(pool, sizeof(*content_info));
			content_info->content_type = oid->oid;
			content_info->content = *data;
		}
	} else {
		/* Assemble a content info with the encap type and data. */
		content_info = PORT_ArenaZAlloc(pool, sizeof(*content_info));
		content_info->content_type = enveloped_data->encrypted_content_info.content_type;
		content_info->content = *data;
	}

	/* Check that we got signed-data out of that. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (!SECITEM_ItemsAreEqual(&oid->oid,
				   &content_info->content_type)) {
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
		if (oid == NULL) {
			pkinit_debug("Internal error.\n");
			return KRB5KRB_ERR_GENERIC;
		}
		oid = SECOID_FindOID(&content_info->content_type);
		if (oid != NULL) {
			pkinit_debug("Reply enveloped data does not contain "
				     "signed data, rather it has \"%s\".\n",
				     oid->desc);
		} else {
			pkinit_debug("Reply enveloped data does not contain "
				     "signed data, but an unknown type.\n");
		}
		return KRB5KRB_ERR_GENERIC;
	}

	/* Decode the signed data. */
	signed_data = bcms_decode_signed_data(pool, &content_info->content);
	if (signed_data == NULL) {
		pkinit_debug("Error parsing signed data part of response.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Verify the signature. */
	signer = NULL;
	status = bcms_verify_signed_data(certdb, signed_data, &signer);
	if (status != SECSuccess) {
		pkinit_debug("Verify error.\n");
		return KRB5KDC_ERR_INVALID_SIG;
	}

	/* Validate the certificate. */
	ret = pkinit_validate_kdc_certificate(certdb, signer,
					      kcontext, request);
	if (ret != 0) {
		pkinit_debug("KDC certificate was not acceptible.\n");
		return ret;
	}

	/* Check that the signed data is what we're expecting. */
	if (!SECITEM_ItemsAreEqual(expected_inner_type,
				   &signed_data->encap_content_info.content_type)) {
		pkinit_debug("Signed data is not PKINIT key data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* That's all for now. */
	*inner_data = &signed_data->encap_content_info.content;
	return 0;
}

/* Verify a draft PA-PK-AS-REP structure, and pull out the interesting bits. */
krb5_error_code
pkinit_verify_draft_pa_pk_as_rep(PLArenaPool *pool,
				 CERTCertDBHandle *certdb,
				 PK11SlotInfo *preferred_slot,
				 struct pkinit_pwcb_args *pwcb_args,
				 krb5_context kcontext,
				 krb5_kdc_req *request,
				 krb5_data *padata,
				 krb5_keyblock *encrypting_key)
{
	struct draft_pa_pk_as_rep *rep;
	SECItem *rkeydata;
	SECStatus status;
	krb5_error_code ret;
	struct content_info *content_info;
	struct draft_reply_key_pack *reply_key_pack;
	SECItem *padata_item;
	SECOidData *oid;
	krb5_keyblock key;
	unsigned long l;

	/* Parse the data (outermost level). */
	padata_item = common_secitem_from_krb5_data(pool, padata);
	rep = pkinit_decode_draft_pa_pk_as_rep(pool, padata_item);
	if (rep == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Check that we know how to handle it. */
	if (rep->draft_pa_pk_as_rep_type !=
	    draft_pa_pk_as_rep_type_enc_key_pack) {
		pkinit_debug("Don't know how to do DH!\n");
		return KRB5KRB_ERR_GENERIC;
	}
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_ENVELOPED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	content_info = &rep->draft_pa_pk_as_rep_value.enc_key_pack;
	if (!SECITEM_ItemsAreEqual(&oid->oid,
				   &content_info->content_type)) {
		pkinit_debug("Reply data is not enveloped data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Extract the reply key data, and validate the KDC. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
	rkeydata = NULL;
	ret = pkinit_verify_enc_key_pack_common(pool,
						certdb,
						preferred_slot,
						pwcb_args,
						kcontext,
						request,
						&content_info->content,
						&oid->oid,
						&rkeydata);
	if (ret != 0) {
		return ret;
	}

	/* Decode the reply_key_pack. */
	reply_key_pack = pkinit_decode_draft_reply_key_pack(pool, rkeydata);
	if (reply_key_pack == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Parse out the key. */
	status = SEC_ASN1DecodeInteger(&reply_key_pack->reply_key.keytype, &l);
	if (status != SECSuccess) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	memset(&key, 0, sizeof(key));
	key.enctype = l;
	key.length = reply_key_pack->reply_key.keyvalue.len;
	key.contents = reply_key_pack->reply_key.keyvalue.data;

	/* Verify the reply_key_pack. */
	status = SEC_ASN1DecodeInteger(&reply_key_pack->nonce, &l);
	if (status != SECSuccess) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (request->nonce != (krb5_int32) l) {
		pkinit_debug("Server verification error.\n");
		return KRB5KRB_AP_ERR_MODIFIED;
	}

	/* Success! */
	return krb5_copy_keyblock_contents(kcontext, &key, encrypting_key);
}

/* Process an enc_key_pack reply, all of it. */
static krb5_error_code
pkinit_process_enc_key_pack(PLArenaPool *pool,
			    CERTCertDBHandle *certdb,
			    PK11SlotInfo *preferred_slot,
			    void *pwcb_args,
			    krb5_context kcontext,
			    krb5_kdc_req *request,
			    krb5_data *request_pkt,
			    struct pa_pk_as_rep *rep,
			    krb5_keyblock *key)
{
	SECStatus status;
	SECOidData *oid;
	const SECItem *roid_data;
	SECItem *rkeydata;
	krb5_checksum checksum;
	krb5_boolean valid;
	krb5_error_code ret;
	struct content_info *content_info;
	struct reply_key_pack *reply_key_pack;
	unsigned long l;

	/* Check that the outer content is an enveloped-data. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_ENVELOPED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	content_info = &rep->pa_pk_as_rep_value.enc_key_pack;
	if (!SECITEM_ItemsAreEqual(&oid->oid,
				   &content_info->content_type)) {
		pkinit_debug("Reply data is not enveloped data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Extract the reply key data, and validate the KDC. */
	roid_data = cert_get_oid_pkinit_rkey_data();
	rkeydata = NULL;
	ret = pkinit_verify_enc_key_pack_common(pool,
						certdb,
						preferred_slot,
						pwcb_args,
						kcontext,
						request,
						&content_info->content,
						roid_data,
						&rkeydata);
	if (ret != 0) {
		return ret;
	}

	/* Decode the reply_key_pack. */
	reply_key_pack = pkinit_decode_reply_key_pack(pool, rkeydata);
	if (reply_key_pack == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Parse out the key. */
	status = SEC_ASN1DecodeInteger(&reply_key_pack->reply_key.keytype, &l);
	if (status != SECSuccess) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	memset(key, 0, sizeof(*key));
	key->enctype = l;
	key->length = reply_key_pack->reply_key.keyvalue.len;
	key->contents = reply_key_pack->reply_key.keyvalue.data;

	/* Verify the reply_key_pack. */
	status = SEC_ASN1DecodeInteger(&reply_key_pack->as_checksum.cksumtype,
				       &l);
	if (status != SECSuccess) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	memset(&checksum, 0, sizeof(checksum));
	checksum.checksum_type = l;
	checksum.length = reply_key_pack->as_checksum.checksum.len;
	checksum.contents = reply_key_pack->as_checksum.checksum.data;
	valid = 0;
	ret = krb5_c_verify_checksum(kcontext, key,
				     KRB5_KEYUSAGE_TGS_REQ_AUTH_CKSUM,
				     request_pkt,
				     &checksum,
				     &valid);
	if ((ret != 0) || (!valid)) {
		pkinit_debug("Server verification error.\n");
		return KRB5KRB_AP_ERR_MODIFIED;
	}

	return 0;
}

/* Process a dh_info reply, all of it. */
static krb5_error_code
pkinit_process_dh_rep_info(PLArenaPool *pool,
			   CERTCertDBHandle *certdb,
			   krb5_context kcontext,
			   krb5_kdc_req *request,
			   struct pa_pk_as_rep *rep,
			   SECKEYPrivateKey *client_dh_privkey,
			   SECKEYPublicKey *client_dh_pubkey,
			   SECItem *client_dh_nonce,
			   krb5_keyblock *key)
{
	SECStatus status;
	SECOidData *oid;
	const SECItem *roid_data;
	SECItem *kdc_pubkey, *kdc_nonce, *shared_secret, spk_bits;
	SECKEYPublicKey server_dh_pubkey;
	PK11SymKey *sym;
	CERTCertificate *kdc_cert;
	krb5_error_code ret;
	struct content_info *content_info;
	struct signed_data *signed_dh_data;
	struct kdc_dh_key_info *kdc_dh_key_info;

	/* Yeah, but only if we actually tried to use DH. */
	if (client_dh_privkey == NULL) {
		pkinit_debug("We didn't generate a DH private key.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (client_dh_pubkey == NULL) {
		pkinit_debug("We didn't generate a DH private key.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Check that the outer content is an enveloped-data. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_SIGNED_DATA);
	if (oid == NULL) {
		pkinit_debug("Internal error.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	content_info = &rep->pa_pk_as_rep_value.dh_info.dh_signed_data;
	if (!SECITEM_ItemsAreEqual(&oid->oid,
				   &content_info->content_type)) {
		pkinit_debug("Reply data is not signed data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Decode the signed-data and verify the KDC's signature. */
	signed_dh_data = bcms_decode_signed_data(pool, &content_info->content);
	if (signed_dh_data == NULL) {
		pkinit_debug("Error parsing KDC's signed data.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	kdc_cert = NULL;
	status = bcms_verify_signed_data(certdb, signed_dh_data, &kdc_cert);
	if (status != SECSuccess) {
		pkinit_debug("Verify error.\n");
		return KRB5KDC_ERR_INVALID_SIG;
	}

	/* Verify that the KDC's certificate is acceptible. */
	ret = pkinit_validate_kdc_certificate(certdb, kdc_cert,
					      kcontext, request);
	if (ret != 0) {
		pkinit_debug("KDC certificate was not acceptible.\n");
		return ret;
	}

	/* Check that the signed data is what we're expecting. */
	roid_data = cert_get_oid_pkinit_dhkey_data();
	content_info = &signed_dh_data->encap_content_info;
	if (!SECITEM_ItemsAreEqual(roid_data,
				   &content_info->content_type)) {
		pkinit_debug("Signed data is not PKINIT DH keying data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Extract the KDC's public key, and derive the shared secret. */
	kdc_dh_key_info = pkinit_decode_kdc_dh_key_info(pool,
							&content_info->content);
	if (kdc_dh_key_info == NULL) {
		pkinit_debug("Error decoding KDC's DH reply data.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Derive the shared secret. */
	spk_bits = kdc_dh_key_info->subject_public_key;
	/* The server's key is a bit string, so its length is a bit count.
	 * Change that back into a byte count. */
	spk_bits.len = (spk_bits.len + 7) / 8;
	kdc_pubkey = common_decode_integer(pool, &spk_bits);
	if (kdc_pubkey == NULL) {
		pkinit_debug("Error decoding KDC's DH public value.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	kdc_nonce = rep->pa_pk_as_rep_value.dh_info.server_dh_nonce;
	server_dh_pubkey = *client_dh_pubkey;
	server_dh_pubkey.u.dh.publicValue = *kdc_pubkey;
	sym = PK11_PubDerive(client_dh_privkey,
			     &server_dh_pubkey,
			     PR_TRUE,
			     NULL,
			     NULL,
			     CKM_DH_PKCS_DERIVE,
			     CKM_TLS_MASTER_KEY_DERIVE_DH,
			     CKA_DERIVE,
			     0,
			     NULL);
	if (sym == NULL) {
		pkinit_debug("Error deriving DH shared secret.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Extract the shared secret. */
	if (PK11_ExtractKeyValue(sym) != SECSuccess) {
		pkinit_debug("Can't extract DH shared secret.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	shared_secret = PK11_GetKeyData(sym);
	if (shared_secret == NULL) {
		pkinit_debug("Error extracting DH shared secret.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Convert the secret to a key. */
	if (pkinit_octet_string_to_aeskey(kcontext, SEC_OID_SHA1,
					  shared_secret,
					  kdc_nonce ?  client_dh_nonce : NULL,
					  kdc_nonce,
					  ENCTYPE_AES256_CTS_HMAC_SHA1_96,
					  key) != SECSuccess) {
		pkinit_debug("Error deriving DH-based encryption key.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Success! */
	return 0;
}

/* Verify a PA-PK-AS-REP structure, and pull out the interesting bits. */
krb5_error_code
pkinit_verify_pa_pk_as_rep(PLArenaPool *pool,
			   CERTCertDBHandle *certdb,
			   PK11SlotInfo *preferred_slot,
			   struct pkinit_pwcb_args *pwcb_args,
			   krb5_context kcontext,
			   krb5_kdc_req *request,
			   krb5_data *request_pkt,
			   krb5_data *padata,
			   SECKEYPrivateKey *client_dh_privkey,
			   SECKEYPublicKey *client_dh_pubkey,
			   SECItem *client_dh_nonce,
			   krb5_keyblock *encrypting_key)
{
	struct pa_pk_as_rep *rep;
	SECItem *padata_item;
	krb5_error_code ret;
	krb5_keyblock key;

	/* Parse the data (outermost level). */
	padata_item = common_secitem_from_krb5_data(pool, padata);
	rep = pkinit_decode_pa_pk_as_rep(pool, padata_item);
	if (rep == NULL) {
		pkinit_debug("Parsing error.\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Check that we know how to handle it, and hand it off to the proper
	 * function. */
	ret = KRB5KRB_ERR_GENERIC;
	switch (rep->pa_pk_as_rep_type) {
	case pa_pk_as_rep_type_enc_key_pack:
		ret = pkinit_process_enc_key_pack(pool,
						  certdb,
						  preferred_slot,
						  pwcb_args,
						  kcontext,
						  request,
						  request_pkt,
						  rep,
						  &key);
		if (ret != 0) {
			return ret;
		}
		break;
	case pa_pk_as_rep_type_dh_info:
		ret = pkinit_process_dh_rep_info(pool,
						 certdb,
						 kcontext,
						 request,
						 rep,
						 client_dh_privkey,
						 client_dh_pubkey,
						 client_dh_nonce,
						 &key);
		if (ret != 0) {
			return ret;
		}
		break;
	default:
		pkinit_debug("Don't know how to do handle KDC reply!\n");
		return KRB5KRB_ERR_GENERIC;
	}

	/* Success! */
	return krb5_copy_keyblock_contents(kcontext, &key, encrypting_key);
}
